/*
 * Copyright (c) 2018 Institute for Integrated Circuits, Johannes Kepler University.

 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:

 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.

 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
*/
package at.jku.iic.pmd;

import java.io.IOException;
import java.io.PrintWriter;
import java.util.HashMap;

import com.microsoft.z3.BoolExpr;
import com.microsoft.z3.Context;
import com.microsoft.z3.IntExpr;
import com.microsoft.z3.IntNum;
import com.microsoft.z3.Model;
import com.microsoft.z3.Solver;
import com.microsoft.z3.Status;
import com.microsoft.z3.Z3Exception;

/**
 * @author bklepic
 *
 */
public class PmdSmtSolver {

    @SuppressWarnings("serial")
    class TestFailedException extends Exception {

        public TestFailedException() {
            super("Check FAILED");
        }
    };
    
    // class Variables (e.g. input Stuff)
    int maxX = 6;	// Grid length
	int maxY = 6;	// Grid high
	int maxT = 10;	// Grid Z (Time)
	int maxNumOfDirChangers = 1;	// max number of direction changes from source to droplet tail
	Context ctx;	// context
	TDependentSample[] samples; // test samples
	Coords[] sources;
	
	public PmdSmtSolver(Context ctx, int maxX, int maxY, int maxT, int maxNumOfDirChangers, TDependentSample[] samples, Coords[] sources) {
		super();
		this.ctx = ctx;
		this.maxX = maxX;
		this.maxY = maxY;
		this.maxT = maxT;
		this.maxNumOfDirChangers = maxNumOfDirChangers;
		this.samples = samples;
		this.sources = sources;
	}
    
	private BoolExpr initVars() {
		BoolExpr initVars = ctx.mkTrue();
		for (TDependentSample sample : samples) {
			for(Sample step : sample.steps){
				for(Coords dirChanger : step.dirChangers){
					initVars = ctx.mkAnd(initVars, ctx.mkOr(
						ctx.mkEq(ctx.mkInt(-2), dirChanger.x), 
						ctx.mkEq(ctx.mkInt(-4), dirChanger.x),
						ctx.mkAnd(ctx.mkLe(ctx.mkInt(0), dirChanger.x), ctx.mkLt(dirChanger.x, ctx.mkInt(maxX)))
					));

					initVars = ctx.mkAnd(initVars, ctx.mkOr(
						ctx.mkEq(ctx.mkInt(-2), dirChanger.y), 
						ctx.mkEq(ctx.mkInt(-4), dirChanger.y),
						ctx.mkAnd(ctx.mkLe(ctx.mkInt(0), dirChanger.y), ctx.mkLt(dirChanger.y, ctx.mkInt(maxY)))
					));
				}
				for(Coords pos : step.sample){
					initVars = ctx.mkAnd(initVars, ctx.mkOr(
							ctx.mkEq(ctx.mkInt(-2), pos.x), 
							ctx.mkEq(ctx.mkInt(-4), pos.x),
							ctx.mkAnd(ctx.mkLe(ctx.mkInt(0), pos.x), ctx.mkLt(pos.x, ctx.mkInt(maxX)))
					));
					initVars = ctx.mkAnd(initVars, ctx.mkOr(
						ctx.mkEq(ctx.mkInt(-2), pos.y), 
						ctx.mkEq(ctx.mkInt(-4), pos.y),
						ctx.mkAnd(ctx.mkLe(ctx.mkInt(0), pos.y), ctx.mkLt(pos.y, ctx.mkInt(maxY)))
					));
				}
				
				initVars = ctx.mkAnd(initVars, ctx.mkOr(
					ctx.mkEq(ctx.mkInt(-2), step.source.x), 
					ctx.mkEq(ctx.mkInt(-4), step.source.x),
					ctx.mkAnd(ctx.mkLe(ctx.mkInt(0), step.source.x), ctx.mkLt(step.source.x, ctx.mkInt(maxX)))
				));
				initVars = ctx.mkAnd(initVars, ctx.mkOr(
					ctx.mkEq(ctx.mkInt(-2), step.source.y), 
					ctx.mkEq(ctx.mkInt(-4), step.source.y),
					ctx.mkAnd(ctx.mkLe(ctx.mkInt(0), step.source.y), ctx.mkLt(step.source.y, ctx.mkInt(maxY)))
				));
			}
		}
		return initVars;
	}
	
	private BoolExpr setStartStopPos() {
		BoolExpr startStop = ctx.mkTrue();
		for (TDependentSample sample : samples) {
			for(Coords dirChanger : sample.steps[sample.steps.length-1].dirChangers){
				startStop = ctx.mkAnd(startStop, ctx.mkEq(dirChanger.x, ctx.mkInt(-2)), ctx.mkEq(dirChanger.y, ctx.mkInt(-2)));					
			}
			startStop = ctx.mkAnd(startStop, ctx.mkEq(sample.steps[sample.steps.length-1].source.x, ctx.mkInt(-2)));
			startStop = ctx.mkAnd(startStop, ctx.mkEq(sample.steps[sample.steps.length-1].source.y, ctx.mkInt(-2)));
			
			for (int i = 0; i < sample.steps[0].sample.length; i++) {
				startStop = ctx.mkAnd(startStop, ctx.mkEq(sample.steps[0].sample[i].x, ctx.mkInt(-2)));
				startStop = ctx.mkAnd(startStop, ctx.mkEq(sample.steps[0].sample[i].y, ctx.mkInt(-2)));
			}
			
			// somewhere we have to reach the startPos (and it have to be before we reach endpos)
			// somewhere we have to reach the endPos
			BoolExpr reachEndPos = ctx.mkFalse();
			BoolExpr reachStartPos = ctx.mkFalse();
			for(Sample step : sample.steps){
				for(Coords s : step.sample){
					reachStartPos = ctx.mkOr(reachStartPos, ctx.mkAnd(
						ctx.mkEq(s.x, sample.startPos.x),
						ctx.mkEq(s.y, sample.startPos.y)
					));
					reachEndPos = ctx.mkOr(reachEndPos, ctx.mkAnd(
						ctx.mkEq(s.x, sample.endPos.x),
						ctx.mkEq(s.y, sample.endPos.y), 
						reachStartPos
					));
				}
			}
			startStop = ctx.mkAnd(startStop, reachEndPos);
			// at the end all sample parts have to be -1
			for (int i = 0; i < sample.steps[sample.steps.length-1].sample.length; i++) {
				startStop = ctx.mkAnd(startStop, ctx.mkEq(sample.steps[sample.steps.length-1].sample[i].x, ctx.mkInt(-4)));
				startStop = ctx.mkAnd(startStop, ctx.mkEq(sample.steps[sample.steps.length-1].sample[i].y, ctx.mkInt(-4)));
			}
		}
		return startStop;
	}
	
	private BoolExpr getMovingConstraint() {
		BoolExpr moving = ctx.mkTrue();
		for (TDependentSample sample : samples) {
			for(int t = 1; t < sample.steps.length; t++){ // start with t=1 so we can check if we moved from step 0->1
				// one of the following 3 must be true for moving constraint
				BoolExpr sameX = ctx.mkEq(sample.steps[t].sample[0].x, sample.steps[t-1].sample[0].x);
				BoolExpr plusOneX = ctx.mkEq(sample.steps[t].sample[0].x, ctx.mkAdd(sample.steps[t-1].sample[0].x, ctx.mkInt(1)));
				BoolExpr minusOneX = ctx.mkEq(sample.steps[t].sample[0].x, ctx.mkSub(sample.steps[t-1].sample[0].x, ctx.mkInt(1)));
				// one of the following 3 must be true for moving constraint
				BoolExpr sameY = ctx.mkEq(sample.steps[t].sample[0].y, sample.steps[t-1].sample[0].y);
				BoolExpr plusOneY = ctx.mkEq(sample.steps[t].sample[0].y, ctx.mkAdd(sample.steps[t-1].sample[0].y, ctx.mkInt(1)));
				BoolExpr minusOneY = ctx.mkEq(sample.steps[t].sample[0].y, ctx.mkSub(sample.steps[t-1].sample[0].y, ctx.mkInt(1)));

				// only one of this 5 can happen at the same time this is due to the mkEqual before
				BoolExpr stay = stay(sample, t, t-1); // stays at its position
				BoolExpr right = ctx.mkAnd(plusOneX, sameY); // move right
				BoolExpr left = ctx.mkAnd(minusOneX, sameY); // move left
				BoolExpr up = ctx.mkAnd(sameX, plusOneY); // move Up
				BoolExpr down = ctx.mkAnd(sameX, minusOneY); // move Down
			
				BoolExpr flowAwayActive = ctx.mkFalse();
				BoolExpr flowAway = ctx.mkTrue();
				for (int i = 0; i < sample.steps[t-1].sample.length; i++) {
					BoolExpr flowAwayActiveTmp =ctx.mkAnd(
						ctx.mkEq(sample.steps[t-1].sample[i].x, sample.endPos.x),
						ctx.mkEq(sample.steps[t-1].sample[i].y, sample.endPos.y)
					);
					flowAwayActive = ctx.mkOr(flowAwayActive, flowAwayActiveTmp);
					flowAway = ctx.mkAnd(flowAway, 
						ctx.mkImplies(
								ctx.mkAnd(ctx.mkNot(stay), flowAwayActiveTmp), 
							ctx.mkAnd(
								ctx.mkEq(sample.steps[t].sample[i].x, ctx.mkInt(-4)),
								ctx.mkEq(sample.steps[t].sample[i].y, ctx.mkInt(-4))
						))
					);
				}
				// and it up all together
				BoolExpr defaultMove = ctx.mkOr(stay, right, left, up, down);

				BoolExpr initSample = ctx.mkAnd(
					ctx.mkEq(sample.steps[t-1].sample[0].x, ctx.mkInt(-2)),
					ctx.mkEq(sample.steps[t-1].sample[0].y, ctx.mkInt(-2))
				);
				BoolExpr initSampleConstraint = ctx.mkImplies(ctx.mkNot(stay),  
					ctx.mkAnd(
						ctx.mkEq(sample.steps[t].sample[0].x, sample.startPos.x),
						ctx.mkEq(sample.steps[t].sample[0].y, sample.startPos.y)
					));
				
				BoolExpr endSampleConstraint = ctx.mkTrue();
				for (int i = 0; i < sample.steps[t-1].sample.length; i++) {
					BoolExpr endSampleConstraintTmp =ctx.mkAnd(
						ctx.mkEq(sample.steps[t-1].sample[i].x, ctx.mkInt(-4)),
						ctx.mkEq(sample.steps[t-1].sample[i].y, ctx.mkInt(-4))
					);
					endSampleConstraint = ctx.mkAnd(endSampleConstraint, 
						ctx.mkImplies(endSampleConstraintTmp, 
							ctx.mkAnd(
								ctx.mkEq(sample.steps[t].sample[i].x, ctx.mkInt(-4)),
								ctx.mkEq(sample.steps[t].sample[i].y, ctx.mkInt(-4))
						))
					);
				}
				
				moving = ctx.mkAnd(moving, endSampleConstraint, 
					(BoolExpr) ctx.mkITE(initSample, initSampleConstraint, 
						(BoolExpr) ctx.mkITE(flowAwayActive, flowAway, defaultMove)
					)
				);
				
				// set previous Variables of sample i.g. [tail, tail+1,tail+2,...,head-2,head-1] 
				BoolExpr samplePartsIfMove = ctx.mkTrue();
				BoolExpr samplePartsIfStay = ctx.mkTrue();
				for(int i = 1; i < sample.steps[t].sample.length; i++){
					// if moving and not staying then every step go 1 ahead..
					samplePartsIfMove = ctx.mkAnd(samplePartsIfMove, 
						ctx.mkEq(sample.steps[t].sample[i].x, sample.steps[t-1].sample[i-1].x),
						ctx.mkEq(sample.steps[t].sample[i].y, sample.steps[t-1].sample[i-1].y)					
					);
					// else.. we stay.. every step is the same like at t-1
					samplePartsIfStay = ctx.mkAnd(samplePartsIfStay, 
						ctx.mkEq(sample.steps[t].sample[i].x, sample.steps[t-1].sample[i].x),
						ctx.mkEq(sample.steps[t].sample[i].y, sample.steps[t-1].sample[i].y)
					);
				}
				moving = ctx.mkAnd(moving, (BoolExpr)ctx.mkITE(stay, samplePartsIfStay, samplePartsIfMove));
				/*
				moving = ctx.mkAnd(moving,
					ctx.mkImplies(stay, samplePartsIfStay),
					ctx.mkImplies(ctx.mkOr(right, left, up, down), samplePartsIfMove)
				);
				*/
			}
		}
		return moving;
	}
	
	private Coords getValidTail(Coords startPos, Coords[] sample, Coords tail){
		IntExpr matchX[] = new IntExpr[sample.length];
		IntExpr matchY[] = new IntExpr[sample.length];
		matchX[sample.length-1] = tail.x;
		matchY[sample.length-1] = tail.y;
		for (int j = sample.length-1; j > 0; j--) {
			matchX[j-1] = (IntExpr) ctx.mkITE(ctx.mkEq(matchX[j], ctx.mkInt(-2)), sample[j-1].x, matchX[j]);
			matchY[j-1] = (IntExpr) ctx.mkITE(ctx.mkEq(matchY[j], ctx.mkInt(-2)), sample[j-1].y, matchY[j]);
		}
		IntExpr retX = (IntExpr) ctx.mkITE(ctx.mkEq(matchX[0], ctx.mkInt(-2)), 
			startPos.x, 
			matchX[0]
		);
		IntExpr retY = (IntExpr) ctx.mkITE(ctx.mkEq(matchY[0], ctx.mkInt(-2)), 
			startPos.y, 
			matchY[0]
		);
		return new Coords(retX, retY);
	}
	
	private BoolExpr stay(TDependentSample sample, int t1, int t2){
		Sample act = sample.steps[t1];
		Sample next = sample.steps[t2];
		BoolExpr stay = ctx.mkTrue();
		for (int i = 0; i < act.sampleLen; i++) {
			stay = ctx.mkAnd(stay, 
					ctx.mkEq(act.sample[i].x, next.sample[i].x), 
					ctx.mkEq(act.sample[i].y, next.sample[i].y)	
			);
		}
		return stay;
	}

	private BoolExpr getSourceConnectionConstraintStep(Sample step, Coords tail, BoolExpr stay){
		// ((dXTail = dirCh1X)OR(dYTail = dirCh1Y) AND 
		// 	(dirCh1X = dirCh2x) OR .. AND 
		// 	(dirCh2x = source0) OR ..)
		BoolExpr pumpConstraintA = ctx.mkOr(
			ctx.mkAnd(
				ctx.mkEq(tail.x, step.dirChangers[0].x),
				ctx.mkLe(tail.y, step.dirChangers[0].y)	
			),
			ctx.mkAnd(
				ctx.mkEq(tail.y, step.dirChangers[0].y),
				ctx.mkLe(tail.x, step.dirChangers[0].x)
			), 
            ctx.mkAnd(
				ctx.mkEq(tail.x, step.dirChangers[0].x),
				ctx.mkGe(tail.y, step.dirChangers[0].y)	
			),
			ctx.mkAnd(
				ctx.mkEq(tail.y, step.dirChangers[0].y),
				ctx.mkGe(tail.x, step.dirChangers[0].x)
			)
		);
		// iterate for every step between and build up pump constraint
		BoolExpr tmpPumpConstraintA = ctx.mkTrue();
		for (int k = 1; k < step.dirChangers.length; k++) {
			tmpPumpConstraintA = ctx.mkAnd(tmpPumpConstraintA, 
				ctx.mkOr(
					ctx.mkAnd(
						ctx.mkEq(step.dirChangers[k-1].x, step.dirChangers[k].x),
						ctx.mkLe(step.dirChangers[k-1].y, step.dirChangers[k].y)	
					),
					ctx.mkAnd(
						ctx.mkEq(step.dirChangers[k-1].y, step.dirChangers[k].y),
						ctx.mkLe(step.dirChangers[k-1].x, step.dirChangers[k].x)
					),
					ctx.mkAnd(
						ctx.mkEq(step.dirChangers[k-1].x, step.dirChangers[k].x),
						ctx.mkGe(step.dirChangers[k-1].y, step.dirChangers[k].y)	
					),
					ctx.mkAnd(
						ctx.mkEq(step.dirChangers[k-1].y, step.dirChangers[k].y),
						ctx.mkGe(step.dirChangers[k-1].x, step.dirChangers[k].x)
					)
				)
			);
		}
		// add it up
		pumpConstraintA = ctx.mkAnd(tmpPumpConstraintA, pumpConstraintA);        
		// OR the sources
		tmpPumpConstraintA = ctx.mkFalse();
		for (int i = 0; i < sources.length; i++) {
			BoolExpr sourceMatchA = ctx.mkOr(
				ctx.mkAnd(
					ctx.mkEq(step.dirChangers[step.dirChangers.length-1].x, sources[i].x),
					ctx.mkLe(step.dirChangers[step.dirChangers.length-1].y, sources[i].y)	
				),
				ctx.mkAnd(
					ctx.mkEq(step.dirChangers[step.dirChangers.length-1].y, sources[i].y),
					ctx.mkLe(step.dirChangers[step.dirChangers.length-1].x, sources[i].x)
				),
				ctx.mkAnd(
					ctx.mkEq(step.dirChangers[step.dirChangers.length-1].x, sources[i].x),
					ctx.mkGe(step.dirChangers[step.dirChangers.length-1].y, sources[i].y)	
				),
				ctx.mkAnd(
					ctx.mkEq(step.dirChangers[step.dirChangers.length-1].y, sources[i].y),
					ctx.mkGe(step.dirChangers[step.dirChangers.length-1].x, sources[i].x)
				)
			);
			
			ctx.mkEq(step.source.x, ctx.mkInt(-2));
			ctx.mkEq(step.source.y, ctx.mkInt(-2));
			tmpPumpConstraintA = ctx.mkOr(tmpPumpConstraintA, 
				ctx.mkAnd(sourceMatchA, 
					ctx.mkEq(step.source.x, sources[i].x), 
					ctx.mkEq(step.source.y, sources[i].y)					
			));
			
		} // for every source
		pumpConstraintA = ctx.mkAnd(pumpConstraintA, tmpPumpConstraintA);
		BoolExpr srcConstraint = ctx.mkAnd(
			(BoolExpr) ctx.mkITE(stay, ctx.mkEq(step.source.x, ctx.mkInt(-2)), ctx.mkNot(ctx.mkEq(step.source.x, ctx.mkInt(-2)))),
			(BoolExpr) ctx.mkITE(stay, ctx.mkEq(step.source.y, ctx.mkInt(-2)), ctx.mkNot(ctx.mkEq(step.source.y, ctx.mkInt(-2))))
		);
        return ctx.mkAnd(srcConstraint, ctx.mkImplies(ctx.mkNot(stay), pumpConstraintA));
	}
	
	private BoolExpr getSourceConnectionConstraintStepFixSrc(Sample step, Coords tail, BoolExpr stay){
		BoolExpr pumpConstraintA = ctx.mkOr(
			ctx.mkAnd(
				ctx.mkEq(tail.x, step.dirChangers[0].x),
				ctx.mkLe(tail.y, step.dirChangers[0].y)	
			),
			ctx.mkAnd(
				ctx.mkEq(tail.y, step.dirChangers[0].y),
				ctx.mkLe(tail.x, step.dirChangers[0].x)
			), 
			 ctx.mkAnd(
				ctx.mkEq(tail.x, step.dirChangers[0].x),
				ctx.mkGe(tail.y, step.dirChangers[0].y)	
			),
			ctx.mkAnd(
				ctx.mkEq(tail.y, step.dirChangers[0].y),
				ctx.mkGe(tail.x, step.dirChangers[0].x)
			)
		);
		
		// iterate for every step between and build up pump constraint
		BoolExpr tmpPumpConstraintA = ctx.mkTrue();
		for (int k = 1; k < step.dirChangers.length; k++) {
			tmpPumpConstraintA = ctx.mkAnd(tmpPumpConstraintA, 
				ctx.mkOr(
					ctx.mkAnd(
						ctx.mkEq(step.dirChangers[k-1].x, step.dirChangers[k].x),
						ctx.mkLe(step.dirChangers[k-1].y, step.dirChangers[k].y)	
					),
					ctx.mkAnd(
						ctx.mkEq(step.dirChangers[k-1].y, step.dirChangers[k].y),
						ctx.mkLe(step.dirChangers[k-1].x, step.dirChangers[k].x)
					), 
					ctx.mkAnd(
						ctx.mkEq(step.dirChangers[k-1].x, step.dirChangers[k].x),
						ctx.mkGe(step.dirChangers[k-1].y, step.dirChangers[k].y)	
					),
					ctx.mkAnd(
						ctx.mkEq(step.dirChangers[k-1].y, step.dirChangers[k].y),
						ctx.mkGe(step.dirChangers[k-1].x, step.dirChangers[k].x)
					)
				)
			);
		}
		// add it up
		pumpConstraintA = ctx.mkAnd(tmpPumpConstraintA, pumpConstraintA);        
		BoolExpr sourceMatchA = ctx.mkOr(
			ctx.mkAnd(
				ctx.mkEq(step.dirChangers[step.dirChangers.length-1].x, step.source.x),//sources[i].x),
				ctx.mkLe(step.dirChangers[step.dirChangers.length-1].y, step.source.y)//sources[i].y)	
			),
			ctx.mkAnd(
				ctx.mkEq(step.dirChangers[step.dirChangers.length-1].y, step.source.y),//sources[i].y),
				ctx.mkLe(step.dirChangers[step.dirChangers.length-1].x, step.source.x)//sources[i].x)
			), 
			ctx.mkAnd(
				ctx.mkEq(step.dirChangers[step.dirChangers.length-1].x, step.source.x),//sources[i].x),
				ctx.mkGe(step.dirChangers[step.dirChangers.length-1].y, step.source.y)//sources[i].y)	
			),
			ctx.mkAnd(
				ctx.mkEq(step.dirChangers[step.dirChangers.length-1].y, step.source.y),//sources[i].y),
				ctx.mkGe(step.dirChangers[step.dirChangers.length-1].x, step.source.x)//sources[i].x)
			)
		);
	
		pumpConstraintA = ctx.mkAnd(pumpConstraintA, sourceMatchA);

		BoolExpr src = ctx.mkAnd(ctx.mkEq(step.source.x, sources[step.id].x),
				ctx.mkEq(step.source.y, sources[step.id].y));
		
        return ctx.mkAnd(src, ctx.mkImplies(ctx.mkNot(stay), pumpConstraintA));
	}
	
	private BoolExpr getSourceConnectionConstraint(){
		BoolExpr sourceConnection = ctx.mkTrue();
		for (TDependentSample sample : samples) {
			for(int t = 1; t < sample.steps.length; t++){
				BoolExpr stay = stay(sample, t, t-1);// stays at its position
				// check if tail already on grid, if yes take it otherwise search for a part which is on the grid
				Coords tail = getValidTail(sample.startPos, sample.steps[t-1].sample, sample.steps[t-1].tail);
				BoolExpr pumpConstraint = getSourceConnectionConstraintStep(sample.steps[t-1], tail, stay);
				//BoolExpr pumpConstraint = ctx.mkEq(sample.steps[t-1].source.x, sources[sample.steps[t-1].id].x);
				//pumpConstraint = ctx.mkAnd(pumpConstraint, ctx.mkEq(sample.steps[t-1].source.y, sources[sample.steps[t-1].id].y));

				// sum up constraints what should be when dirChangers are not needed or needed (-1)	
				BoolExpr dirChangerNotUsed = ctx.mkTrue();
				BoolExpr dirChangerUsed = ctx.mkTrue();
				for (Coords dirChanger : sample.steps[t-1].dirChangers) {
					dirChangerNotUsed = ctx.mkAnd(dirChangerNotUsed, 
						ctx.mkEq(dirChanger.x, ctx.mkInt(-2)), 
						ctx.mkEq(dirChanger.y, ctx.mkInt(-2))
					);
					dirChangerUsed = ctx.mkAnd(dirChangerUsed, 
						ctx.mkNot(ctx.mkEq(dirChanger.x, ctx.mkInt(-2))), 
						ctx.mkNot(ctx.mkEq(dirChanger.y, ctx.mkInt(-2))),
						ctx.mkNot(ctx.mkEq(dirChanger.x, ctx.mkInt(-4))), 
						ctx.mkNot(ctx.mkEq(dirChanger.y, ctx.mkInt(-4)))
					);
				}
				// when move (not staying) then constraint of Pump have to be fulfilled
				// when staying the dirChangers should be marked as unused somehow
				sourceConnection = ctx.mkAnd(sourceConnection, pumpConstraint, (BoolExpr) ctx.mkITE(stay, dirChangerNotUsed, dirChangerUsed));
			}
		}
		return sourceConnection;
	}
	
	private BoolExpr getLineCrossingConstraint(BoolExpr stayA, BoolExpr stayB, Coords line1P1, Coords line1P2, Coords line2P1, Coords line2P2){
		
		BoolExpr matchX1 = ctx.mkEq(line1P1.x, line1P2.x);
		BoolExpr matchX2 = ctx.mkEq(line2P1.x, line2P2.x);
		BoolExpr matchY1 = ctx.mkEq(line1P1.y, line1P2.y);
		BoolExpr matchY2 = ctx.mkEq(line2P1.y, line2P2.y);	
		
		BoolExpr matchXX = ctx.mkAnd(matchX1, matchX2); // both lines have are going along the y axis
		BoolExpr matchYY = ctx.mkAnd(matchY1, matchY2); // both lines have are going along the x axis
		BoolExpr matchXY = ctx.mkAnd(matchX1, matchY2); // line 1 is going along y and line 2 along x axis
		BoolExpr matchYX = ctx.mkAnd(matchY1, matchX2); // line 1 is going along x and line 2 along y axis
		
		BoolExpr y1LtY2, y1GtY2, x1LtX2, x1GtX2;
		y1LtY2 = ctx.mkAnd(
				ctx.mkLt(line2P1.y, line1P1.y), 
				ctx.mkLt(line2P2.y, line1P1.y),
				ctx.mkLt(line2P1.y, line1P2.y),
				ctx.mkLt(line2P2.y, line1P2.y)
		);
		y1GtY2 = ctx.mkAnd(
				ctx.mkGt(line2P1.y, line1P1.y), 
				ctx.mkGt(line2P2.y, line1P1.y),
				ctx.mkGt(line2P1.y, line1P2.y),
				ctx.mkGt(line2P2.y, line1P2.y)
		);
		x1LtX2 = ctx.mkAnd(
				ctx.mkLt(line2P1.x, line1P1.x), 
				ctx.mkLt(line2P2.x, line1P1.x),
				ctx.mkLt(line2P1.x, line1P2.x),
				ctx.mkLt(line2P2.x, line1P2.x)
		);
		x1GtX2 = ctx.mkAnd(
				ctx.mkGt(line2P1.x, line1P1.x), 
				ctx.mkGt(line2P2.x, line1P1.x),
				ctx.mkGt(line2P1.x, line1P2.x),
				ctx.mkGt(line2P2.x, line1P2.x)
		);
		// matchXX
		// if X==X check Y had to be smaller or greater
		BoolExpr xEqX = ctx.mkEq(line1P1.x, line2P1.x); // if both lines are going along y axis and have the same x coordinate
		BoolExpr noCrossXX = ctx.mkImplies(xEqX, ctx.mkOr(y1LtY2, y1GtY2)); // if both on same x then y should not cross 
		// matchYY
		// if Y==Y check X had to be smaller or greater
		BoolExpr yEqY = ctx.mkEq(line1P1.y, line2P1.y); // if both lines are going along x axis and have the same y coordinate
		BoolExpr noCrossYY = ctx.mkImplies(yEqY, ctx.mkOr(x1LtX2, x1GtX2)); // if both on same y then x should not cross 
		// matchXY
		// if line1 is placed along the y axis and line 2 along the x axis
		BoolExpr noCrossXY = ctx.mkImplies(ctx.mkNot(ctx.mkOr(y1LtY2, y1GtY2)), ctx.mkOr(x1LtX2, x1GtX2)); // if line2 Y is between line1 Y1-Y2 then X of both should not cross 
		// matchYX
		// if line1 is placed along the y axis and line 2 along the x axis
		BoolExpr noCrossYX = ctx.mkImplies(ctx.mkNot(ctx.mkOr(x1LtX2, x1GtX2)), ctx.mkOr(y1LtY2, y1GtY2)); // if line2 X is between line1 X1-X2 then Y of both should not cross 
		
		return ctx.mkImplies(ctx.mkNot(ctx.mkOr(stayA, stayB)), 
				ctx.mkAnd(
						ctx.mkImplies(matchXX, noCrossXX), 
						ctx.mkImplies(matchYY, noCrossYY), 
						ctx.mkImplies(matchXY, noCrossXY), 
						ctx.mkImplies(matchYX, noCrossYX)));
	}
	
	private BoolExpr getSinglePointToPointCrossingConstraint(Sample sampleA, Sample sampleB){
		BoolExpr toReturn = ctx.mkTrue();
		
		toReturn = ctx.mkOr(
			ctx.mkAnd(
					ctx.mkEq(sampleA.head.x, ctx.mkInt(-2)),
					ctx.mkEq(sampleA.head.y, ctx.mkInt(-2)),
					ctx.mkEq(sampleB.head.x, ctx.mkInt(-2)),
					ctx.mkEq(sampleB.head.y, ctx.mkInt(-2))
				),
				ctx.mkAnd(
					ctx.mkEq(sampleA.head.x, ctx.mkInt(-4)),
					ctx.mkEq(sampleA.head.y, ctx.mkInt(-4)),
					ctx.mkEq(sampleB.head.x, ctx.mkInt(-4)),
					ctx.mkEq(sampleB.head.y, ctx.mkInt(-4))
				),
				ctx.mkNot(ctx.mkAnd(
					ctx.mkEq(sampleA.head.x, sampleB.head.x),
					ctx.mkEq(sampleA.head.y, sampleB.head.y)
				))
			);
//		
//		
//		
//		for (int i = 0; i < sampleA.sampleLen; i++) {
//			for (int j = 0; j < sampleB.sampleLen; j++) {
//				toReturn = ctx.mkAnd(toReturn, ctx.mkOr(
//					ctx.mkAnd(
//						ctx.mkEq(sampleA.sample[i].x, ctx.mkInt(-2)),
//						ctx.mkEq(sampleA.sample[i].y, ctx.mkInt(-2)),
//						ctx.mkEq(sampleB.sample[j].x, ctx.mkInt(-2)),
//						ctx.mkEq(sampleB.sample[j].y, ctx.mkInt(-2))
//					),
//					ctx.mkAnd(
//						ctx.mkEq(sampleA.sample[i].x, ctx.mkInt(-4)),
//						ctx.mkEq(sampleA.sample[i].y, ctx.mkInt(-4)),
//						ctx.mkEq(sampleB.sample[j].x, ctx.mkInt(-4)),
//						ctx.mkEq(sampleB.sample[j].y, ctx.mkInt(-4))
//					),
//					ctx.mkNot(ctx.mkAnd(
//						ctx.mkEq(sampleA.sample[i].x, sampleB.sample[j].x),
//						ctx.mkEq(sampleA.sample[i].y, sampleB.sample[j].y)
//					))
//				));
//			}
//		}
		return toReturn;
	}
	
	/**
	 * returns the constraint that a source to target line should not cross any other non moving samples 
	 */
	private BoolExpr getSinglePointToLineCrossingConstraint(BoolExpr stay, Sample sample, Coords line1P1, Coords line1P2){
		BoolExpr matchX1 = ctx.mkEq(line1P1.x, line1P2.x);
		BoolExpr matchY1 = ctx.mkEq(line1P1.y, line1P2.y);	
		BoolExpr toReturn = ctx.mkTrue();
		for (int i = 0; i < sample.sampleLen; i++) {
			BoolExpr matchXX = ctx.mkAnd(matchX1, ctx.mkEq(line1P1.x, sample.sample[i].x));
			BoolExpr matchYY = ctx.mkAnd(matchY1, ctx.mkEq(line1P1.y, sample.sample[i].y));
			
			BoolExpr xxCond = ctx.mkImplies(matchXX, ctx.mkOr(
				ctx.mkAnd(
					ctx.mkGt(sample.sample[i].y, line1P1.y),
					ctx.mkGt(sample.sample[i].y, line1P2.y)
				),
				ctx.mkAnd(
					ctx.mkLt(sample.sample[i].y, line1P1.y),
					ctx.mkLt(sample.sample[i].y, line1P2.y)
				)
			));
			BoolExpr yyCond = ctx.mkImplies(matchYY, ctx.mkOr(
				ctx.mkAnd(
					ctx.mkGt(sample.sample[i].x, line1P1.x),
					ctx.mkGt(sample.sample[i].x, line1P2.x)
				),
				ctx.mkAnd(
					ctx.mkLt(sample.sample[i].x, line1P1.x),
					ctx.mkLt(sample.sample[i].x, line1P2.x)
				)
			));
			// only valid if sample part != -1 (exist already on grid)
			toReturn = ctx.mkAnd(toReturn, ctx.mkImplies(
					ctx.mkAnd(
						ctx.mkNot(ctx.mkOr(
							ctx.mkEq(sample.sample[i].x, ctx.mkInt(-2)),
							ctx.mkEq(sample.sample[i].y, ctx.mkInt(-2)),
							ctx.mkEq(sample.sample[i].x, ctx.mkInt(-4)),
							ctx.mkEq(sample.sample[i].y, ctx.mkInt(-4))
						)),
						ctx.mkNot(stay)
					),
				ctx.mkAnd(xxCond, yyCond)));			
		}
		return toReturn;
	}
	
	private BoolExpr getLineCrossingConstraint(){
		BoolExpr crossingLines = ctx.mkTrue();
		for(int t = 1; t < maxT; t++){
			for (int i = 0; i < samples.length; i++) {
				BoolExpr stayA = stay(samples[i], t, t-1);// stays at its position
				Coords tailA = getValidTail(samples[i].startPos, samples[i].steps[t-1].sample, samples[i].steps[t-1].tail);
				for (int i2 = i + 1; i2 < samples.length; i2++) {
					BoolExpr stayB = stay(samples[i2], t, t-1);// stays at its position
					Coords tailB = getValidTail(samples[i2].startPos, samples[i2].steps[t-1].sample, samples[i2].steps[t-1].tail);
					BoolExpr crossLineToLine = ctx.mkTrue();
					//---------------------------------------------------------------------------------------
					// the sample parts tail to head and future flow parts should not be crossed by other samples tail to head and future flow path
					BoolExpr crossSampleSS = ctx.mkTrue();
					// only check head positions cause rest is checked implicit TODO
					for(int futuret1 = t-1; futuret1 < maxT; futuret1++){
						for(int futuret2 = t-1; futuret2 < maxT; futuret2++){
							crossSampleSS = ctx.mkAnd(crossSampleSS, ctx.mkImplies(ctx.mkNot(ctx.mkOr(stayA, stayB)),  
								getSinglePointToPointCrossingConstraint(samples[i].steps[futuret1], samples[i2].steps[futuret2])));
						}
					}
					// the sample parts tail to head and future flow parts should not be crossed by line 1 of other sample
					BoolExpr crossSampleS1 = ctx.mkTrue();
					for(int future = t-1; future < maxT; future++){
						crossSampleS1 = ctx.mkAnd(crossSampleS1, ctx.mkImplies(ctx.mkNot(stayA),
							getSinglePointToLineCrossingConstraint(stayB, samples[i].steps[future], 
								tailB, samples[i2].steps[t-1].dirChangers[0])));
					}
					// the sample parts tail to head and future flow parts should not be crossed by any other Line of other sample
					BoolExpr crossSampleSN = ctx.mkTrue();
					for (int j = 1; j < samples[i2].steps[t-1].dirChangers.length; j++) {
						for(int future = t-1; future < maxT; future++){
							crossSampleSN = ctx.mkAnd(crossSampleSN, ctx.mkImplies(ctx.mkNot(stayA),
								getSinglePointToLineCrossingConstraint(stayB, samples[i].steps[future], 
									samples[i2].steps[t-1].dirChangers[j-1], samples[i2].steps[t-1].dirChangers[j])));
						}
					}
					// the sample parts tail to head and future flow parts should not be crossed by last line of other sample (last DirCh and Source)
					int lastPosL = samples[i2].steps[t-1].dirChangers.length-1;
					BoolExpr crossSampleSL = ctx.mkTrue();
					for(int future = t-1; future < maxT; future++){
						crossSampleSL = ctx.mkAnd(crossSampleSL, ctx.mkImplies(ctx.mkNot(stayA),
							getSinglePointToLineCrossingConstraint(stayB, samples[i].steps[future], 
								samples[i2].steps[t-1].dirChangers[lastPosL], samples[i2].steps[t-1].source)));
					}
					// add it all up for tail to head of A
					crossLineToLine = ctx.mkAnd(crossLineToLine, crossSampleSS, crossSampleS1, crossSampleSN, crossSampleSL);
					//---------------------------------------------------------------------------------------
					// line 1 should not cross tail to head and future flow path of other sample
					BoolExpr crossSample1S = ctx.mkTrue();
					for(int future = t-1; future < maxT; future++){
						crossSample1S = ctx.mkAnd(crossSample1S, ctx.mkImplies(ctx.mkNot(stayA),
							getSinglePointToLineCrossingConstraint(stayB, samples[i2].steps[future], 
								tailA, samples[i].steps[t-1].dirChangers[0])));
					}
					// line 1 should not cross line 1 of other sample
					BoolExpr crossLine11 = getLineCrossingConstraint(stayA, stayB, 
						tailA, samples[i].steps[t-1].dirChangers[0], 
						tailB, samples[i2].steps[t-1].dirChangers[0]
					);
					// Line 1 should not cross any other Line of other sample
					BoolExpr crossLine1N = ctx.mkTrue();
					for (int j = 1; j < samples[i2].steps[t-1].dirChangers.length; j++) {
						crossLine1N = ctx.mkAnd(crossLine1N, 
								getLineCrossingConstraint(stayA, stayB, 
								tailA, samples[i].steps[t-1].dirChangers[0], 
								samples[i2].steps[t-1].dirChangers[j-1], samples[i2].steps[t-1].dirChangers[j]
							));
					} 
					// Line 1 should not cross last line of other sample (last DirCh and Source)
					int lastPos1L = samples[i2].steps[t-1].dirChangers.length-1;
					BoolExpr crossLine1L = getLineCrossingConstraint(stayA, stayB, 
							tailA, samples[i].steps[t-1].dirChangers[0], 
							samples[i2].steps[t-1].dirChangers[lastPos1L], samples[i2].steps[t-1].source);
					// add it all up for Line 1 of A
					crossLineToLine = ctx.mkAnd(crossLineToLine, crossSample1S, crossLine11, crossLine1N, crossLine1L);
					//---------------------------------------------------------------------------------------
					//Line 2 to Line N-1 of this sample should not cross Line 1-N of other sample
					for (int n = 1; n < samples[i].steps[t-1].dirChangers.length; n++) {
						// line N should not cross tail to head and future flow path of other sample
						BoolExpr crossSampleNS = ctx.mkTrue();
						for(int future = t-1; future < maxT; future++){
							crossSampleNS = ctx.mkAnd(crossSampleNS, ctx.mkImplies(ctx.mkNot(stayB),
								getSinglePointToLineCrossingConstraint(stayA, samples[i2].steps[future], 
									samples[i].steps[t-1].dirChangers[n-1], samples[i].steps[t-1].dirChangers[n])));
						}
						// line N should not cross line 1 of other sample
						BoolExpr crossLineN1 = getLineCrossingConstraint(stayA, stayB,  
							samples[i].steps[t-1].dirChangers[n-1], samples[i].steps[t-1].dirChangers[n],
							tailB, samples[i2].steps[t-1].dirChangers[0]
						);
						// Line N should not cross any other Line of other sample
						BoolExpr crossLineNN = ctx.mkTrue();
						for (int j = 1; j < samples[i2].steps[t-1].dirChangers.length; j++) {
							crossLineNN = ctx.mkAnd(crossLineNN, 
									getLineCrossingConstraint(stayA, stayB,  
									samples[i].steps[t-1].dirChangers[n-1], samples[i].steps[t-1].dirChangers[n], 
									samples[i2].steps[t-1].dirChangers[j-1], samples[i2].steps[t-1].dirChangers[j]
								));
						}
						// Line N should not cross last line of other sample (last DirCh and Source)
						int lastPosNL = samples[i2].steps[t-1].dirChangers.length-1;
						BoolExpr crossLineNL = getLineCrossingConstraint(stayA, stayB, 
								samples[i].steps[t-1].dirChangers[n-1], samples[i].steps[t-1].dirChangers[n],
								samples[i2].steps[t-1].dirChangers[lastPosNL], samples[i2].steps[t-1].source);
						// add it all up for Line N of A
						crossLineToLine = ctx.mkAnd(crossLineToLine, crossSampleNS, crossLineN1, crossLineNN, crossLineNL);
					}
					//---------------------------------------------------------------------------------------
					// last line should not cross tail to head and future flow path of other sample
					BoolExpr crossSampleLS = ctx.mkTrue();
					int lastPosLS = samples[i].steps[t-1].dirChangers.length-1;
					for(int future = t-1; future < maxT; future++){
						crossSampleLS = ctx.mkAnd(crossSampleLS, ctx.mkImplies(ctx.mkNot(stayB),
							getSinglePointToLineCrossingConstraint(stayA, samples[i2].steps[future], 
									samples[i].steps[t-1].dirChangers[lastPosLS], samples[i].steps[t-1].source)));
					}
					// last line of this sample should not cross line 1 of other sample
					BoolExpr crossLineL1 = getLineCrossingConstraint(stayA, stayB,  
						samples[i].steps[t-1].dirChangers[lastPosLS], samples[i].steps[t-1].source,
						tailB, samples[i2].steps[t-1].dirChangers[0]
					);
					// last line should not cross any other Line of other sample
					BoolExpr crossLineLN = ctx.mkTrue();
					for (int j = 1; j < samples[i2].steps[t-1].dirChangers.length; j++) {
						crossLineLN = ctx.mkAnd(crossLineLN, 
								getLineCrossingConstraint(stayA, stayB,  
								samples[i].steps[t-1].dirChangers[lastPosLS], samples[i].steps[t-1].source,
								samples[i2].steps[t-1].dirChangers[j-1], samples[i2].steps[t-1].dirChangers[j]
							));
					}
					// last line should not cross last line of other sample (last DirCh and Source)
					BoolExpr crossLineLL = getLineCrossingConstraint(stayA, stayB,  
						samples[i].steps[t-1].dirChangers[lastPosLS], samples[i].steps[t-1].source,
						samples[i2].steps[t-1].dirChangers[samples[i2].steps[t-1].dirChangers.length-1], samples[i2].steps[t-1].source
					);
					// add it all up for Line L of A
					crossLineToLine = ctx.mkAnd(crossLineToLine, crossSampleLS, crossLineL1, crossLineLN, crossLineLL);
					//---------------------------------------------------------------------------------------					
					// only need the constraint to not cross this specific line if My Line and the specific other line are moving (handled already in sub-function)
					crossingLines = ctx.mkAnd(crossingLines, ctx.mkImplies(ctx.mkNot(ctx.mkOr(stayA, stayB)), crossLineToLine));
				}
			}
		}
		return crossingLines;
	}
	
	private BoolExpr getPointToLineCrossingConstraint(){
		BoolExpr crossingLines = ctx.mkTrue();
		for(int t = 1; t < maxT; t++){
			for (int i = 0; i < samples.length; i++) {
				BoolExpr stayA = stay(samples[i], t, t-1);// stays at its position
				Coords tailA = getValidTail(samples[i].startPos, samples[i].steps[t-1].sample, samples[i].steps[t-1].tail);
				for (int i2 = 0; i2 < samples.length; i2++) {
					if(i != i2){
						BoolExpr crossLineToLine = ctx.mkTrue();
						//---------------------------------------------------------------------------------------
						// the sample parts tail to head and future flow parts should not be crossed by other sample
						BoolExpr crossSampleS = ctx.mkTrue();
						for(int futuret1 = t-1; futuret1 < maxT; futuret1++){
							crossSampleS = ctx.mkAnd(crossSampleS, ctx.mkImplies(ctx.mkNot(stayA),  
								getSinglePointToPointCrossingConstraint(samples[i].steps[futuret1], samples[i2].steps[t-1])));
						}
						crossLineToLine = ctx.mkAnd(crossLineToLine, crossSampleS);
						// line 1 of this sample should not be crossed by other sample
						BoolExpr crossLine1 = getSinglePointToLineCrossingConstraint(stayA, samples[i2].steps[t-1], tailA, samples[i].steps[t-1].dirChangers[0]);
						crossLineToLine = ctx.mkAnd(crossLineToLine, crossLine1);
						// Line 2.. N-1 of this sample should not be crossed by other sample
						for (int n = 1; n < samples[i].steps[t-1].dirChangers.length; n++) {
							// line N should not cross other sample
							BoolExpr crossLineN = getSinglePointToLineCrossingConstraint(stayA, samples[i2].steps[t-1], samples[i].steps[t-1].dirChangers[n-1], samples[i].steps[t-1].dirChangers[n]);
							crossLineToLine = ctx.mkAnd(crossLineToLine, crossLineN);
						}
						// last line should not cross other sample
						BoolExpr crossLineL = getSinglePointToLineCrossingConstraint(stayA, samples[i2].steps[t-1], 
								samples[i].steps[t-1].dirChangers[samples[i].steps[t-1].dirChangers.length-1], samples[i].steps[t-1].source);
						crossLineToLine = ctx.mkAnd(crossLineToLine, crossLineL);
						//---------------------------------------------------------------------------------------					
						// only need the constraint to not cross this specific line if My Line is moving and the specific other line not moving or moving (dont care)
						crossingLines = ctx.mkAnd(crossingLines, ctx.mkImplies(ctx.mkNot(stayA), crossLineToLine));
					}
				}
			}
		}
		return crossingLines;
	}
	
	private BoolExpr getCrossingSampleConstraint(){
		BoolExpr crossing = ctx.mkTrue();
		/*
		 * Samples should not cross each other
		 * 	a) There should be no other Sample between Sample Head And Tail
		 *  b) and also this sample should not cross between itself
		 */
		for (int i = 0; i < samples.length; i++) {
			/* should also not cross itself in case the sample is so long that it could move in a circle so j=i
			 * 	and also cause moving a step back is not allowed
			 */
			for(Sample sample : samples[i].steps){
				for (int j = 0; j < sample.sample.length; j++) {
					for (int j2 = j+1; j2 < sample.sample.length; j2++) {
						crossing = ctx.mkAnd(crossing, ctx.mkOr(
							ctx.mkAnd(
								ctx.mkEq(sample.sample[j].x, ctx.mkInt(-2)),
								ctx.mkEq(sample.sample[j].y, ctx.mkInt(-2))
							),
							ctx.mkAnd(
								ctx.mkEq(sample.sample[j].x, ctx.mkInt(-4)),
								ctx.mkEq(sample.sample[j].y, ctx.mkInt(-4))
							),
							ctx.mkNot(ctx.mkAnd(
								ctx.mkEq(sample.sample[j].x, sample.sample[j2].x),
								ctx.mkEq(sample.sample[j].y, sample.sample[j2].y)
							))
						));
					}
				}
			}
			for (int j = i+1; j < samples.length; j++) { 
				int maxSteps = samples[i].numTimeSteps > samples[j].numTimeSteps ? samples[j].numTimeSteps : samples[i].numTimeSteps; 
				for(int t = 0; t < maxSteps; t++){
					BoolExpr check = ctx.mkTrue();
					// no other sample(headToTail) should cross the way of this sample(headToTail)
					for(Coords sample1 : samples[i].steps[t].sample){
						for(Coords sample2 : samples[j].steps[t].sample){
							check = ctx.mkAnd(check, ctx.mkOr(
								ctx.mkAnd(
									ctx.mkEq(sample1.x, ctx.mkInt(-2)),
									ctx.mkEq(sample1.y, ctx.mkInt(-2)),
									ctx.mkEq(sample2.x, ctx.mkInt(-2)),
									ctx.mkEq(sample2.y, ctx.mkInt(-2))
								),
								ctx.mkAnd(
									ctx.mkEq(sample1.x, ctx.mkInt(-4)),
									ctx.mkEq(sample1.y, ctx.mkInt(-4)),
									ctx.mkEq(sample2.x, ctx.mkInt(-4)),
									ctx.mkEq(sample2.y, ctx.mkInt(-4))
								),
								ctx.mkNot(ctx.mkAnd(
									ctx.mkEq(sample1.x, sample2.x),
									ctx.mkEq(sample1.y, sample2.y)
								))
							));
						}
					}
					crossing = ctx.mkAnd(crossing, check);
				}
			}		
		}
		return crossing;
	}
	
	private BoolExpr getCrossingConstraintSameLine(BoolExpr stayA, Coords line1P1, Coords line1P2, Coords line2P1, Coords line2P2){
		BoolExpr matchX1 = ctx.mkEq(line1P1.x, line1P2.x);
		BoolExpr matchX2 = ctx.mkEq(line2P1.x, line2P2.x);
		BoolExpr matchY1 = ctx.mkEq(line1P1.y, line1P2.y);
		BoolExpr matchY2 = ctx.mkEq(line2P1.y, line2P2.y);	
		
		BoolExpr matchXX = ctx.mkAnd(matchX1, matchX2); // both lines have are going along the y axis
		BoolExpr matchYY = ctx.mkAnd(matchY1, matchY2); // both lines have are going along the x axis
		BoolExpr matchXY = ctx.mkAnd(matchX1, matchY2); // line 1 is going along y and line 2 along x axis
		BoolExpr matchYX = ctx.mkAnd(matchY1, matchX2); // line 1 is going along x and line 2 along y axis
		
		BoolExpr y1LtY2, y1GtY2, x1LtX2, x1GtX2;
		
		y1LtY2 = ctx.mkAnd(
			ctx.mkLe(line2P1.y, line1P1.y), 
			ctx.mkLe(line2P2.y, line1P1.y),
			ctx.mkLe(line2P1.y, line1P2.y),
			ctx.mkLe(line2P2.y, line1P2.y)
		);
		y1GtY2 = ctx.mkAnd(
			ctx.mkGe(line2P1.y, line1P1.y), 
			ctx.mkGe(line2P2.y, line1P1.y),
			ctx.mkGe(line2P1.y, line1P2.y),
			ctx.mkGe(line2P2.y, line1P2.y)
		);
		x1LtX2 = ctx.mkAnd(
			ctx.mkLe(line2P1.x, line1P1.x), 
			ctx.mkLe(line2P2.x, line1P1.x),
			ctx.mkLe(line2P1.x, line1P2.x),
			ctx.mkLe(line2P2.x, line1P2.x)
		);
		x1GtX2 = ctx.mkAnd(
			ctx.mkGe(line2P1.x, line1P1.x), 
			ctx.mkGe(line2P2.x, line1P1.x),
			ctx.mkGe(line2P1.x, line1P2.x),
			ctx.mkGe(line2P2.x, line1P2.x)
		);
		
		// matchXX
		// if X==X check Y had to be smaller or greater
		BoolExpr xEqX = ctx.mkEq(line1P1.x, line2P1.x); // if both lines are going along y axis and have the same x coordinate
		BoolExpr noCrossXX = ctx.mkImplies(xEqX, ctx.mkOr(y1LtY2, y1GtY2)); // if both on same x then y should not cross 
		// matchYY
		// if Y==Y check X had to be smaller or greater
		BoolExpr yEqY = ctx.mkEq(line1P1.y, line2P1.y); // if both lines are going along x axis and have the same y coordinate
		BoolExpr noCrossYY = ctx.mkImplies(yEqY, ctx.mkOr(x1LtX2, x1GtX2)); // if both on same y then x should not cross 
		// matchXY
		// if line1 is placed along the y axis and line 2 along the x axis
		BoolExpr noCrossXY = ctx.mkImplies(ctx.mkNot(ctx.mkOr(y1LtY2, y1GtY2)), ctx.mkOr(x1LtX2, x1GtX2)); // if line2 Y is between line1 Y1-Y2 then X of both should not cross 
		// matchYX
		// if line1 is placed along the y axis and line 2 along the x axis
		BoolExpr noCrossYX = ctx.mkImplies(ctx.mkNot(ctx.mkOr(x1LtX2, x1GtX2)), ctx.mkOr(y1LtY2, y1GtY2)); // if line2 X is between line1 X1-X2 then Y of both should not cross 
		
		return ctx.mkImplies(ctx.mkNot(stayA), 
				ctx.mkAnd(
						ctx.mkImplies(matchXX, noCrossXX), 
						ctx.mkImplies(matchYY, noCrossYY), 
						ctx.mkImplies(matchXY, noCrossXY), 
						ctx.mkImplies(matchYX, noCrossYX)));
	}
	
	/**
	 * returns the constraint that a source to target line should not cross any other non moving samples 
	 */
	private BoolExpr getSameSinglePointToLineCrossingConstraint(Sample sample, Coords line1P1, Coords line1P2){
		BoolExpr matchX1 = ctx.mkEq(line1P1.x, line1P2.x);
		BoolExpr matchY1 = ctx.mkEq(line1P1.y, line1P2.y);	
		BoolExpr toReturn = ctx.mkTrue();
		for (int i = 0; i < sample.sampleLen; i++) {
			BoolExpr matchXX = ctx.mkAnd(matchX1, ctx.mkEq(line1P1.x, sample.sample[i].x));
			BoolExpr matchYY = ctx.mkAnd(matchY1, ctx.mkEq(line1P1.y, sample.sample[i].y));
			
			BoolExpr xxCond = ctx.mkImplies(matchXX, ctx.mkOr(
				ctx.mkAnd(
					ctx.mkGe(sample.sample[i].y, line1P1.y),
					ctx.mkGe(sample.sample[i].y, line1P2.y)
				),
				ctx.mkAnd(
					ctx.mkLe(sample.sample[i].y, line1P1.y),
					ctx.mkLe(sample.sample[i].y, line1P2.y)
				)
			));
			BoolExpr yyCond = ctx.mkImplies(matchYY, ctx.mkOr(
				ctx.mkAnd(
					ctx.mkGe(sample.sample[i].x, line1P1.x),
					ctx.mkGe(sample.sample[i].x, line1P2.x)
				),
				ctx.mkAnd(
					ctx.mkLe(sample.sample[i].x, line1P1.x),
					ctx.mkLe(sample.sample[i].x, line1P2.x)
				)
			));
			
			// only valid if sample part != -1 (exist already on grid)
			toReturn = ctx.mkAnd(toReturn, ctx.mkImplies(
					ctx.mkNot(ctx.mkOr(
							ctx.mkEq(sample.sample[i].x, ctx.mkInt(-2)),
							ctx.mkEq(sample.sample[i].y, ctx.mkInt(-2)),
							ctx.mkEq(sample.sample[i].x, ctx.mkInt(-4)),
							ctx.mkEq(sample.sample[i].y, ctx.mkInt(-4))
						)),
				ctx.mkAnd(xxCond, yyCond)));			
		}
		return toReturn;
	}
	
	private BoolExpr getSameLineCrossingConstraint(){
		BoolExpr crossingLines = ctx.mkTrue();
		for(int t = 1; t < maxT; t++){
			for (int i = 0; i < samples.length; i++) {
				BoolExpr stayA = stay(samples[i], t, t-1);// stays at its position
				Coords tailA = getValidTail(samples[i].startPos, samples[i].steps[t-1].sample, samples[i].steps[t-1].tail);
				BoolExpr crossLineToLine = ctx.mkTrue();
				//---------------------------------------------------------------------------------------
				// the sample parts tail to head and future flow parts should not be crossed by line 1 of other sample
				BoolExpr crossSampleS1 = ctx.mkTrue();
				for(int future = t-1; future < maxT; future++){
					crossSampleS1 = ctx.mkAnd(crossSampleS1, 
						getSameSinglePointToLineCrossingConstraint(samples[i].steps[future], 
							tailA, samples[i].steps[t-1].dirChangers[0]));
				}
				// the sample parts tail to head and future flow parts should not be crossed by any other Line of other sample
				BoolExpr crossSampleSN = ctx.mkTrue();
				for (int j = 1; j < samples[i].steps[t-1].dirChangers.length; j++) {
					for(int future = t-1; future < maxT; future++){
						crossSampleSN = ctx.mkAnd(crossSampleSN, 
							getSameSinglePointToLineCrossingConstraint(samples[i].steps[future], 
								samples[i].steps[t-1].dirChangers[j-1], samples[i].steps[t-1].dirChangers[j]));
					}
				}
				// the sample parts tail to head and future flow parts should not be crossed by last line of other sample (last DirCh and Source)
				int lastPosL = samples[i].steps[t-1].dirChangers.length-1;
				BoolExpr crossSampleSL = ctx.mkTrue();
				for(int future = t-1; future < maxT; future++){
					crossSampleSL = ctx.mkAnd(crossSampleSL, 
						getSameSinglePointToLineCrossingConstraint(samples[i].steps[future], 
							samples[i].steps[t-1].dirChangers[lastPosL], samples[i].steps[t-1].source));
				}
				// add it all up 
				crossLineToLine = ctx.mkAnd(crossLineToLine, crossSampleS1, crossSampleSN, crossSampleSL);
				//---------------------------------------------------------------------------------------
				// Line 1 should not cross any other Line of this sample
				BoolExpr crossLine1N = ctx.mkTrue();
				for (int j = 1; j < samples[i].steps[t-1].dirChangers.length; j++) {
					crossLine1N = ctx.mkAnd(crossLine1N, 
						getCrossingConstraintSameLine(stayA, 
							tailA, samples[i].steps[t-1].dirChangers[0], 
							samples[i].steps[t-1].dirChangers[j-1], samples[i].steps[t-1].dirChangers[j]
					));
				} 
				// Line 1 should not cross last line of this sample
				int lastPos1L = samples[i].steps[t-1].dirChangers.length-1;
				BoolExpr crossLine1L = getCrossingConstraintSameLine(stayA, 
					tailA, samples[i].steps[t-1].dirChangers[0], 
					samples[i].steps[t-1].dirChangers[lastPos1L], samples[i].steps[t-1].source
				);
				crossLineToLine = ctx.mkAnd(crossLineToLine, crossLine1N, crossLine1L);
				//---------------------------------------------------------------------------------------
				//Line 2..N-1 of this sample should not cross Line 1..N of other sample
				for (int n = 1; n < samples[i].steps[t-1].dirChangers.length; n++) {
					// Line n should not cross any other Line of this sample
					BoolExpr crossLineNN = ctx.mkTrue();
					for (int j = n + 1; j < samples[i].steps[t-1].dirChangers.length; j++) {
						crossLineNN = ctx.mkAnd(crossLineNN, 
							getCrossingConstraintSameLine(stayA,  
								samples[i].steps[t-1].dirChangers[n-1], samples[i].steps[t-1].dirChangers[n], 
								samples[i].steps[t-1].dirChangers[j-1], samples[i].steps[t-1].dirChangers[j]
						));
					}
					// Line n should not cross last line of this sample
					int lastPosNL = samples[i].steps[t-1].dirChangers.length-1;
					BoolExpr crossLineNL = getCrossingConstraintSameLine(stayA,
						samples[i].steps[t-1].dirChangers[n-1], samples[i].steps[t-1].dirChangers[n],
						samples[i].steps[t-1].dirChangers[lastPosNL], samples[i].steps[t-1].source
					);
					crossLineToLine = ctx.mkAnd(crossLineToLine, crossLineNN, crossLineNL);
				}
				//---------------------------------------------------------------------------------------					
				// only need the constraint to not cross this specific line if My Line and the specific other line are moving
				crossingLines = ctx.mkAnd(crossingLines, crossLineToLine);
			}
		}
		return crossingLines;
	}
	
	private void printSolution(Model m, PrintWriter writer) {
		int[][][][] grid = new int[maxX][maxY][maxT][samples.length];
        for (int i = 0; i < maxX; i++) {
            for (int j = 0; j < maxY; j++) {
                for (int k = 0; k < maxT; k++) {
                	for (int l = 0; l < samples.length; l++) {
                		grid[i][j][k][l] = 0;
                	}
                }
            }
        }
        for (int t = 0; t < maxT; t++) {
        	for (int l = 0; l < samples.length; l++) {
        		int[] x = new int[samples[l].steps[t].sample.length];
        		int[] y = new int[samples[l].steps[t].sample.length];
        		for (int i = 0; i < x.length; i++){
        			x[i] = ((IntNum) m.evaluate(samples[l].steps[t].sample[i].x, false)).getInt();
        			y[i] = ((IntNum) m.evaluate(samples[l].steps[t].sample[i].y, false)).getInt();
        			if((x[i] != -2 && y[i] != -2) && (x[i] != -4 && y[i] != -4)){
            			grid[x[i]][y[i]][t][l] = 1;
        			}

				}
        	}
        }
        
        // print grid:
        writer.println("Solution:");
        for (int t = 0; t < maxT; t++) {
        	for (int k = 0; k < samples.length; k++) {
        		writer.print("S" + (k+1) + ": ");
        		writer.print("dest[" + 
        				((IntNum) samples[k].endPos.x).getInt() + "|" + 
        				((IntNum) samples[k].endPos.y).getInt() + 
        				"] ");
        		writer.print("head[" + 
        				((IntNum) m.evaluate(samples[k].steps[t].head.x, false)).getInt() + "|" + 
        				((IntNum) m.evaluate(samples[k].steps[t].head.y, false)).getInt() + 
        				"] ");
        		int matchX[] = new int[samples[k].steps[t].sample.length];
        		int matchY[] = new int[samples[k].steps[t].sample.length];
        		matchX[samples[k].steps[t].sample.length-1] = ((IntNum) m.evaluate(samples[k].steps[t].tail.x, false)).getInt();
        		matchY[samples[k].steps[t].sample.length-1] = ((IntNum) m.evaluate(samples[k].steps[t].tail.y, false)).getInt();
        		for (int j = samples[k].steps[t].sample.length-1; j > 0; j--) {
        			matchX[j-1] = matchX[j] == -1 ? ((IntNum) m.evaluate(samples[k].steps[t].sample[j-1].x, false)).getInt() : matchX[j];
        			matchY[j-1] = matchY[j] == -1 ? ((IntNum) m.evaluate(samples[k].steps[t].sample[j-1].y, false)).getInt() : matchY[j];
        		}
                writer.print("tail[" + matchX[0] + "|" + matchY[0] + "] ");
//                writer.print("tail[" + 
//        				((IntNum) m.evaluate(samples[k].steps[t].tail.x, false)).getInt() + "|" + 
//        				((IntNum) m.evaluate(samples[k].steps[t].tail.y, false)).getInt() + 
//        				"] ");
                writer.print("DirCh");
        		for (Coords dirChanger : samples[k].steps[t].dirChangers) {
                	writer.print("[" +
            			((IntNum) m.evaluate(dirChanger.x, false)).getInt()
            			+ "|" + 
            			((IntNum) m.evaluate(dirChanger.y, false)).getInt()
            			+ "] "
            		);
				}
        		writer.print("Src[" +
            			((IntNum) m.evaluate(samples[k].steps[t].source.x, false)).getInt()
            			+ "|" + 
            			((IntNum) m.evaluate(samples[k].steps[t].source.y, false)).getInt()
            			+ "] ");
            	writer.println();
        	}
            for (int j = maxY - 1; j >= 0; j--) {
                for (int k = 0; k < maxX; k++) {
                	String num = "0";
                	for (int l = 0; l < samples.length; l++) {
                    	if(grid[k][j][t][l] == 1){
                    		if(num.equals("0")){
                    			num = String.valueOf(1+l);
                    		} else{
                        		num = "x";
                    		}
                    	}
                	}
                	writer.print(num + " ");
                }
                writer.println();
            }
            writer.println();
        }
	}
	
	/**
     * @param solutionPath 
	 * @throws TestFailedException
	 * @throws IOException 
     */
	long[] solve(String solutionPath) throws TestFailedException, IOException {
		PrintWriter writer = new PrintWriter(solutionPath, "UTF-8");	
		//-----------------------------------------------------------
		long[] toReturn = {0, 0, 0, 0};
		System.out.println("init...");
		toReturn[0] = System.currentTimeMillis();
		
		System.out.println("...initVars");
		long time = System.currentTimeMillis();
		BoolExpr initVars = initVars();
		System.out.println(System.currentTimeMillis() - time);
		
		System.out.println("...startStop");
		time = System.currentTimeMillis();
		BoolExpr startStop = setStartStopPos();
		System.out.println(System.currentTimeMillis() - time);
		
		System.out.println("...moving");
		time = System.currentTimeMillis();
		BoolExpr moving = getMovingConstraint();
		System.out.println(System.currentTimeMillis() - time);
		
		System.out.println("...sourceConnection");
		time = System.currentTimeMillis();
		BoolExpr sourceConnection = getSourceConnectionConstraint();
		System.out.println(System.currentTimeMillis() - time);
		
		System.out.println("...crossingSamples");
		time = System.currentTimeMillis();
		BoolExpr crossingSamples = getCrossingSampleConstraint();
		System.out.println(System.currentTimeMillis() - time);
		
		System.out.println("...crossingLines");
		time = System.currentTimeMillis();
		BoolExpr crossingLines = getLineCrossingConstraint();
		System.out.println(System.currentTimeMillis() - time);
		
		System.out.println("...crossingSameLine");
		time = System.currentTimeMillis();
		BoolExpr crossingSameLine = getSameLineCrossingConstraint();
		System.out.println(System.currentTimeMillis() - time);
		
		System.out.println("...pointLine");
		time = System.currentTimeMillis();
		BoolExpr pointLine = getPointToLineCrossingConstraint();
		System.out.println(System.currentTimeMillis() - time);
		
		// add further constraints here... 
		Solver s = ctx.mkSolver();
        s.add(ctx.mkAnd(initVars, startStop, moving,
        		sourceConnection,
        		crossingSamples,
        		crossingLines,
        		crossingSameLine,
        		pointLine
		));
        toReturn[0] = System.currentTimeMillis()-toReturn[0];
		//-----------------------------------------------------------
//        System.out.println("write code to File...");
//        try{
//        	BoolExpr toFile = ctx.mkAnd(initVars, startStop, moving, sourceConnection, crossingSamples, crossingLines, crossingSameLine, pointLine);
//			PrintWriter initCode = new PrintWriter("constraints", "UTF-8");
//			initCode.println(toFile.toString());
//			initCode.close();
//		} catch (IOException e) {
//			// do something
//			System.err.println("IOException");
//			e.printStackTrace();
//		}
		//-----------------------------------------------------------
		System.out.println("solve...");
        toReturn[1] = System.currentTimeMillis();
        Status stat = s.check();
        toReturn[1] = System.currentTimeMillis()-toReturn[1];
		System.out.println("write to file...");
		//-----------------------------------------------------------
        if (stat == Status.SATISFIABLE) {
        	toReturn[3] = 1;
        	toReturn[2] = System.currentTimeMillis();
        	Model m = s.getModel();
        	printSolution(m, writer);
        	toReturn[2] = System.currentTimeMillis()-toReturn[2];
        }else{
        	toReturn[3] = 0;
            writer.println("Failed to solve");
            writer.println(new TestFailedException().getStackTrace());
        }
        writer.close();
		System.out.println("done...");
        return toReturn;
    }
	
	public static void main(String[] args) {
        try {
            // turn on warning messages
            com.microsoft.z3.Global.ToggleWarningMessages(true);
            
            // This example need model generation turned on.
            HashMap<String, String> cfg = new HashMap<String, String>();
            cfg.put("model", "true");
            
           
            String cfgPath = "data.cfg";
            String solutionPath = "solution.txt";
            String resultPath = "result.out";
            Parser p;
            if(args.length == 3){
            	cfgPath = args[0];
            	solutionPath = args[1];
            	resultPath = args[2];
            }
        	p = new Parser(cfgPath);
            
            Context ctx = new Context(cfg);
            
    		TDependentSample[] samples;
    		samples = new TDependentSample[p.fieldCfg.flownum];
    		Coords start, end;
    		for (int i = 0; i < samples.length; i++) {
        		start = new Coords(ctx.mkInt(p.fieldCfg.flows[i].sx), ctx.mkInt(p.fieldCfg.flows[i].sy));
        		end = new Coords(ctx.mkInt(p.fieldCfg.flows[i].ex), ctx.mkInt(p.fieldCfg.flows[i].ey));
        		samples[i] = new TDependentSample(ctx, p.fieldCfg.flows[i].length, p.fieldCfg.flows[i].dirchnum, p.fieldCfg.boundz, start, end);
			}
    		Coords[] sources = new Coords[p.fieldCfg.srcnum];
    		for (int i = 0; i < sources.length; i++) {
        		sources[i] = new Coords(ctx.mkInt(p.fieldCfg.sources[i].srcx), ctx.mkInt(p.fieldCfg.sources[i].srcy));
			}
            PmdSmtSolver solver = new PmdSmtSolver(ctx, p.fieldCfg.boundx, p.fieldCfg.boundy, p.fieldCfg.boundz, p.fieldCfg.dirchnum, samples, sources);
            
            long[] retVal = solver.solve(solutionPath);
            PrintWriter writer = new PrintWriter(resultPath, "UTF-8");
            writer.println("[result]");
            writer.println("inputFile=" + cfgPath);
        	writer.println("solutionFile=" + solutionPath);
            writer.println("TimeToSetUp=" + retVal[0]);
            writer.println("TimeToSolve=" + retVal[1]);
            writer.println("TimeToPrint=" + retVal[2]);
        	writer.println("Satisfiable=" + retVal[3]);
            writer.close();
          
            System.exit(0);
        } catch (Z3Exception ex) {
            System.err.println("Z3 Managed Exception: " + ex.getMessage());
            System.err.println("Stack trace: ");
            ex.printStackTrace(System.err);
            System.exit(-1);
        } catch (TestFailedException ex) {
            System.err.println("TEST CASE FAILED: " + ex.getMessage());
            System.err.println("Stack trace: ");
            ex.printStackTrace(System.err);
            System.exit(-1);
        } catch (Exception ex) {
            System.out.println("Unknown Exception: " + ex.getMessage());
            System.out.println("Stack trace: ");
            ex.printStackTrace(System.out);
            System.exit(-1);
        }
    }

}
