/*
 * Copyright (c) 2018 Institute for Integrated Circuits, Johannes Kepler University.

 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:

 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.

 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
*/
package at.jku.iic.pmd;

import java.awt.Point;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;
import java.util.stream.Collectors;

/**
 * Represents the current state of the grid on which the samples have to be routed.
 */
public class Grid {

    /**
     * Represents the time points until when a grid cell is blocked.
     */
    private final int[][] grid;
    private final List<Point> inputs;
    private final List<Point> exits;

    /**
     * A constant representing that a grid cell is occupied. Must be negative.
     */
    private static final int OCCUPIED = -1;

    public Grid(int size, List<Point> obstacles, List<Point> inputs, List<Point> exits, List<Sample> samples) {

        grid = new int[size + 2][size + 2];
        for (Point p : obstacles) {
            grid[p.y][p.x] = OCCUPIED;
        }

        for (int i = 0; i < grid.length; i++) {
            if (!inputs.contains(new Point(0, i)) && !exits.contains(new Point(0, i))) grid[i][0] = OCCUPIED;
            if (!inputs.contains(new Point(i, grid.length - 1)) && !exits.contains(new Point(i, grid.length - 1)))
                grid[grid.length - 1][i] = OCCUPIED;
            if (!inputs.contains(new Point(grid.length - 1, i)) && !exits.contains(new Point(grid.length - 1, i)))
                grid[i][grid.length - 1] = OCCUPIED;
            if (!inputs.contains(new Point(i, 0)) && !exits.contains(new Point(i, 0))) grid[0][i] = OCCUPIED;
        }

        this.inputs = inputs;
        this.exits = exits;

        // Init grid with starting positions of samples
        for (Sample s : samples) {
            if (s.routeType == RouteType.GRID_TO_EXIT || s.routeType == RouteType.GRID_TO_GRID) {
                for (Point p : s.startPos) {
                    grid[p.y][p.x] = OCCUPIED;
                }
            }
        }
    }

    public int getNextTimeOfGridChange(int time) {

        int min = -1;
        for (int i = 1; i < grid.length - 1; i++) {
            for (int j = 1; j < grid[i].length - 1; j++) {
                if (grid[i][j] >= time && (min == -1 || min > grid[i][j])) {
                    min = grid[i][j];
                }
            }
        }
        return min == -1 ? -1 : min + 1;
    }

    public int getLatestTimeOfGridChange() {

        int max = -1;
        for (int i = 0; i < grid.length; i++) {
            for (int j = 0; j < grid[i].length; j++) {
                if (grid[i][j] >= max) {
                    max = grid[i][j];
                }
            }
        }
        return max;
    }

    public List<Point> getExits() {

        return Collections.unmodifiableList(exits);
    }

    public List<Point> getInputs() {

        return Collections.unmodifiableList(inputs);
    }

    private Set<Point> getObstacles(int time) {

        Set<Point> obstacles = new HashSet<>();
        for (int i = 0; i < grid.length; i++) {
            for (int j = 0; j < grid[i].length; j++) {
                if (!checkAvailability(new Point(j, i), time)) {
                    obstacles.add(new Point(j, i));
                }
            }
        }
        return obstacles;
    }

    public String getStringRep(Sample s, int time, Point usedInput, Point usedExit) {

        StringBuilder sb = new StringBuilder();

        // Size of the grid
        sb.append(grid[0].length).append(' ').append(grid.length).append('\n');

        Set<Point> obstacles = getObstacles(time);

        Set<Point> unusedInputs = new HashSet<Point>(this.inputs);
        unusedInputs.remove(usedInput);
        obstacles.addAll(unusedInputs);

        Set<Point> unusedExits = new HashSet<Point>(this.exits);
        unusedExits.remove(usedExit);
        obstacles.addAll(unusedExits);

        if (s != null) {

            switch (s.routeType) {
            case INPUT_TO_EXIT:
                break;
            case INPUT_TO_GRID:
                for (int i = 1; i < s.endPos.size() - 1; i++) {
                    obstacles.add(new Point(s.endPos.get(i)));
                }
                break;
            case GRID_TO_EXIT:
                obstacles.remove(s.startPos.get(0));
                obstacles.remove(s.startPos.get(s.startPos.size() - 1));
                break;
            default:
                assert s.routeType == RouteType.GRID_TO_GRID;
                obstacles.remove(s.startPos.get(0));
                obstacles.remove(s.startPos.get(s.startPos.size() - 1));

                for (int i = 1; i < s.endPos.size() - 1; i++) {
                    obstacles.add(new Point(s.endPos.get(i)));
                }
                break;
            }
        }
        sb.append(obstacles.size()).append('\n');
        sb.append(obstacles.stream().map(i -> i.x + " " + i.y).collect(Collectors.joining("\n")));
        return sb.toString();
    }

    public boolean checkAvailability(List<Point> points, int time) {

        return points.stream().allMatch(p -> checkAvailability(p, time));
    }

    public boolean checkAvailability(Point p, int time) {

        return grid[p.y][p.x] != OCCUPIED && grid[p.y][p.x] <= time;
    }

    public void updateGrid(Sample s, int time, String result) {

        List<Point> first = new LinkedList<>();
        List<Point> second = new LinkedList<>();
        List<Point> third = new LinkedList<>();

        int i = 0;
        for (String strI : result.split("\n")) {
            int j = 0;
            for (String strJ : strI.split(" ")) {
                if (strJ.startsWith("S") || strJ.startsWith("w") || strJ.startsWith("T")) {
                    if (strJ.endsWith("0")) first.add(new Point(j, i));
                    else if (strJ.endsWith("1")) second.add(new Point(j, i));
                    else if (strJ.endsWith("2")) third.add(new Point(j, i));
                }
                j++;
            }
            i++;
        }

        int blockedUntil;
        switch (s.routeType) {
        case INPUT_TO_EXIT:
            assert second.size() == 0 && third.size() == 0;
            blockedUntil = time + s.sampleLength + first.size() - 2;
            break;
        case INPUT_TO_GRID:
            assert third.size() == 0;
            blockedUntil = time + s.sampleLength + first.size() - 2;
            break;
        case GRID_TO_EXIT:
            assert third.size() == 0;
            blockedUntil = time + s.sampleLength + second.size() - 2;
            break;
        default:
            assert s.routeType == RouteType.GRID_TO_GRID;
            blockedUntil = time + s.sampleLength + second.size() - 2;
            break;
        }

        // Block the path until the calculated time
        List<Point> allPoints = new ArrayList<>();
        allPoints.addAll(first);
        allPoints.addAll(second);
        allPoints.addAll(third);
        for (Point p : allPoints) {
            grid[p.y][p.x] = blockedUntil;
        }

        // Block the target positions of the samples
        switch (s.routeType) {
        case INPUT_TO_GRID:
        case GRID_TO_GRID:
            for (Point p : s.endPos) {
                grid[p.y][p.x] = OCCUPIED;
            }
            break;
        default:
            break;
        }

        // Free the source positions of the samples
        switch (s.routeType) {
        case GRID_TO_EXIT:
        case GRID_TO_GRID:
            for (Point p : s.startPos) {
                grid[p.y][p.x] = blockedUntil;
            }
            break;
        default:
            break;
        }
    }

    public String toString(int time) {

        StringBuilder sb = new StringBuilder();
        for (int i = 0; i < grid.length; i++) {
            for (int j = 0; j < grid.length; j++) {
                if (checkAvailability(new Point(j, i), time)) sb.append(String.format("%2d", 0)).append(' ');
                else sb.append(String.format("%2d", grid[i][j])).append(' ');
            }
            sb.append('\n');
        }
        return sb.toString();
    }

    public static List<Point> getPointsInRectangle(Point topLeft, Point bottomRight) {

        List<Point> points = new LinkedList<>();
        for (int i = topLeft.y; i <= bottomRight.y; i++) {
            for (int j = topLeft.x; j <= bottomRight.x; j++) {
                points.add(new Point(j, i));
            }
        }
        return points;
    }
}
