/*
 * Copyright (c) 2018 Institute for Integrated Circuits, Johannes Kepler University.

 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:

 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.

 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
*/
package at.jku.iic.pmd;

import java.awt.Point;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.LinkedList;
import java.util.List;

/**
 * Represents the source and target position of a sample.
 */
public class Sample {

    public final RouteType routeType;
    public final int sampleLength;
    public final Point input;
    public final Point exit;
    public final List<Point> startPos;
    public final List<Point> endPos;
    public final int routingSegments;

    private Sample(RouteType routeType, int sampleLength, Point input, Point exit, List<Point> startPos, List<Point> endPos,
            int routingSegments) {
        super();
        this.routeType = routeType;
        this.sampleLength = sampleLength;
        this.input = input;
        this.exit = exit;
        this.startPos = startPos;
        this.endPos = endPos;
        this.routingSegments = routingSegments;
    }

    public static Sample createInputToExitSample(int sampleLength, Point input, Point exit) {

        return new Sample(RouteType.INPUT_TO_EXIT, sampleLength, input, exit, null, null, 1);
    }

    public static Sample createInputToGridSample(int sampleLength, Point input, List<Point> endPos) {

        if (sampleLength < 2)
            throw new UnsupportedOperationException("Samples must have a length of at least 2! A sample length of 1 is unsupported.");
        return new Sample(RouteType.INPUT_TO_GRID, sampleLength, input, null, null, endPos, 2);
    }

    public static Sample createGridToExitSample(int sampleLength, Point exit, List<Point> startPos) {

        if (sampleLength < 2)
            throw new UnsupportedOperationException("Samples must have a length of at least 2! A sample length of 1 is unsupported.");
        return new Sample(RouteType.GRID_TO_EXIT, sampleLength, null, exit, startPos, null, 2);
    }

    public static Sample createGridToGridSample(int sampleLength, List<Point> startPos, List<Point> endPos) {

        if (sampleLength < 2)
            throw new UnsupportedOperationException("Samples must have a length of at least 2! A sample length of 1 is unsupported.");
        return new Sample(RouteType.GRID_TO_GRID, sampleLength, null, null, startPos, endPos, 3);
    }

    /**
     * Determines a list of possible end points for the required routes. This is a list because often
     * different inputs and exits can be used.
     * 
     * @param grid
     *            The current state of the grid.
     * @param time
     *            The current time point.
     * @return Returns a list with possible routings.
     */
    public List<RoutingDescription> getRoutingDescriptions(Grid grid, int time) {

        List<RoutingDescription> routingDescriptions = new LinkedList<>();

        switch (routeType) {
        case INPUT_TO_EXIT: {
            if (grid.checkAvailability(input, time) && grid.checkAvailability(exit, time)) {
                StringBuilder sb = new StringBuilder();
                sb.append(routingSegments).append('\n');
                sb.append(2).append(' ').append(input.x).append(' ').append(input.y).append(' ').append(exit.x).append(' ').append(exit.y)
                        .append('\n');
                routingDescriptions.add(new RoutingDescription(input, exit, sb.toString()));
            }
            break;
        }
        case INPUT_TO_GRID: {
            Point tail = endPos.get(endPos.size() - 1);
            Point head = endPos.get(0);

            List<Point> sortedExits = new ArrayList<>(grid.getExits());
            Collections.sort(sortedExits, new ClosestSorter(head));

            for (Point exit : sortedExits) {
                if (grid.checkAvailability(input, time) && grid.checkAvailability(exit, time) && grid.checkAvailability(endPos, time)) {
                    StringBuilder sb = new StringBuilder();
                    sb.append(routingSegments).append('\n');
                    sb.append(2).append(' ').append(input.x).append(' ').append(input.y).append(' ').append(tail.x).append(' ')
                            .append(tail.y).append('\n');
                    sb.append(2).append(' ').append(head.x).append(' ').append(head.y).append(' ').append(exit.x).append(' ').append(exit.y)
                            .append('\n');
                    routingDescriptions.add(new RoutingDescription(input, exit, sb.toString()));
                }
            }
            break;
        }
        case GRID_TO_EXIT: {
            Point tail = startPos.get(startPos.size() - 1);
            Point head = startPos.get(0);

            List<Point> sortedInputs = new ArrayList<>(grid.getInputs());
            Collections.sort(sortedInputs, new ClosestSorter(tail));

            for (Point input : sortedInputs) {
                if (grid.checkAvailability(input, time) && grid.checkAvailability(exit, time)) {
                    StringBuilder sb = new StringBuilder();
                    sb.append(routingSegments).append('\n');
                    sb.append(2).append(' ').append(input.x).append(' ').append(input.y).append(' ').append(tail.x).append(' ')
                            .append(tail.y).append('\n');
                    sb.append(2).append(' ').append(head.x).append(' ').append(head.y).append(' ').append(exit.x).append(' ').append(exit.y)
                            .append('\n');
                    routingDescriptions.add(new RoutingDescription(input, exit, sb.toString()));
                }
            }
            break;
        }
        default: {
            assert routeType == RouteType.GRID_TO_GRID;
            Point tailStart = startPos.get(startPos.size() - 1);
            Point headStart = startPos.get(0);
            Point tailEnd = endPos.get(endPos.size() - 1);
            Point headEnd = endPos.get(0);

            List<Point> sortedExits = new ArrayList<>(grid.getExits());
            Collections.sort(sortedExits, new ClosestSorter(headEnd));

            List<Point> sortedInputs = new ArrayList<>(grid.getInputs());
            Collections.sort(sortedInputs, new ClosestSorter(tailStart));

            for (Point input : sortedInputs) {
                for (Point exit : sortedExits) {

                    if (grid.checkAvailability(input, time) && grid.checkAvailability(exit, time) && grid.checkAvailability(endPos, time)) {
                        StringBuilder sb = new StringBuilder();
                        sb.append(routingSegments).append('\n');

                        sb.append(2).append(' ').append(input.x).append(' ').append(input.y).append(' ').append(tailStart.x).append(' ')
                                .append(tailStart.y).append('\n');

                        sb.append(2).append(' ').append(headStart.x).append(' ').append(headStart.y).append(' ').append(tailEnd.x)
                                .append(' ').append(tailEnd.y).append('\n');

                        sb.append(2).append(' ').append(headEnd.x).append(' ').append(headEnd.y).append(' ').append(exit.x).append(' ')
                                .append(exit.y).append('\n');

                        routingDescriptions.add(new RoutingDescription(input, exit, sb.toString()));
                    }
                }
            }
            break;
        }
        }

        return routingDescriptions;
    }

    private class ClosestSorter implements Comparator<Point> {

        private final Point head;

        public ClosestSorter(Point head) {
            this.head = head;
        }

        @Override
        public int compare(Point boarderPos1, Point boarderPos2) {

            return getDistance(boarderPos1) - getDistance(boarderPos2);
        }

        private int getDistance(Point p) {

            return Math.abs(head.x - p.x) + Math.abs(head.y - p.y);
        }
    }
}
