/*
 * Copyright (c) 2018 Institute for Integrated Circuits, Johannes Kepler University.

 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:

 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.

 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
*/
package at.jku.iic.pmd.benchmark;

import java.awt.Point;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;

import at.jku.iic.pmd.TaskContainer;
import at.jku.iic.pmd.Grid;
import at.jku.iic.pmd.Sample;

/**
 * A hard coded IVD-benchmark. This benchmark was taken from the paper &quot;High-level synthesis of digital
 * microfluidic biochips&quot; and has been modified in order to be suitable for a PMD. For further details we
 * refer to our paper &quot;Sound Valve-Control for Programmable Microfluidic Devices&quot;.
 */
public class InVitro1 {

    public static List<TaskContainer> getBenchmark() {

        return Arrays.asList(getBio1(), getBio2(), getBio3(), getBio4(), getBio5(), getBio6_a(), getBio6_b(), getBio7(), getBio8(),
                getBio9(), getBio10(), getBio11());
    }

    public static TaskContainer getBio1() {

        List<Point> obstacles = Collections.emptyList();
        List<Point> inputs = Arrays.asList(new Point(1, 0), new Point(16, 17));
        List<Point> exits = Arrays.asList(new Point(16, 0), new Point(1, 17));

        Sample s1 = Sample.createGridToGridSample(2, Arrays.asList(new Point(1, 5), new Point(1, 4)),
                Arrays.asList(new Point(6, 10), new Point(6, 9)));
        Sample s2 = Sample.createGridToGridSample(2, Arrays.asList(new Point(16, 5), new Point(16, 4)),
                Arrays.asList(new Point(7, 10), new Point(7, 9)));
        Sample s3 = Sample.createGridToGridSample(2, Arrays.asList(new Point(5, 1), new Point(4, 1)),
                Arrays.asList(new Point(13, 10), new Point(12, 10)));
        Sample s4 = Sample.createGridToGridSample(2, Arrays.asList(new Point(12, 1), new Point(11, 1)),
                Arrays.asList(new Point(12, 16), new Point(13, 16)));

        List<Sample> samples = Arrays.asList(s1, s2, s3, s4);
        Grid grid = new Grid(16, obstacles, inputs, exits, samples);

        return new TaskContainer(grid, samples);
    }

    public static TaskContainer getBio2() {

        List<Point> obstacles = Grid.getPointsInRectangle(new Point(7, 9), new Point(10, 12));
        obstacles.addAll(Grid.getPointsInRectangle(new Point(12, 11), new Point(15, 15)));

        List<Point> inputs = Arrays.asList(new Point(1, 0), new Point(16, 17));
        List<Point> exits = Arrays.asList(new Point(16, 0), new Point(1, 17));

        Sample s1 = Sample.createGridToGridSample(2, Arrays.asList(new Point(1, 5), new Point(1, 4)),
                Arrays.asList(new Point(2, 11), new Point(2, 10)));
        Sample s2 = Sample.createGridToGridSample(2, Arrays.asList(new Point(1, 12), new Point(1, 11)),
                Arrays.asList(new Point(6, 4), new Point(6, 5)));
        Sample s3 = Sample.createGridToGridSample(2, Arrays.asList(new Point(16, 12), new Point(16, 13)),
                Arrays.asList(new Point(7, 4), new Point(7, 5)));
        Sample s4 = Sample.createGridToGridSample(2, Arrays.asList(new Point(12, 1), new Point(13, 1)),
                Arrays.asList(new Point(3, 4), new Point(4, 4)));

        List<Sample> samples = Arrays.asList(s1, s2, s3, s4);
        Grid grid = new Grid(16, obstacles, inputs, exits, samples);

        return new TaskContainer(grid, samples);
    }

    public static TaskContainer getBio3() {

        List<Point> obstacles = Grid.getPointsInRectangle(new Point(7, 3), new Point(10, 6));

        List<Point> inputs = Arrays.asList(new Point(1, 0), new Point(16, 17));
        List<Point> exits = Arrays.asList(new Point(16, 0), new Point(1, 17));

        Sample s1 = Sample.createGridToGridSample(2, Arrays.asList(new Point(2, 11), new Point(2, 10)),
                Arrays.asList(new Point(1, 14), new Point(1, 13)));
        Sample s2 = Sample.createGridToGridSample(2, Arrays.asList(new Point(16, 12), new Point(16, 11)),
                Arrays.asList(new Point(2, 14), new Point(2, 13)));
        Sample s3 = Sample.createGridToGridSample(2, Arrays.asList(new Point(11, 10), new Point(12, 10)),
                Arrays.asList(new Point(6, 9), new Point(7, 9)));
        Sample s4 = Sample.createGridToGridSample(2, Arrays.asList(new Point(5, 11), new Point(5, 10)),
                Arrays.asList(new Point(9, 14), new Point(9, 13)));
        Sample s5 = Sample.createGridToGridSample(2, Arrays.asList(new Point(16, 5), new Point(16, 4)),
                Arrays.asList(new Point(10, 14), new Point(10, 13)));
        Sample s6 = Sample.createGridToGridSample(2, Arrays.asList(new Point(12, 16), new Point(13, 16)),
                Arrays.asList(new Point(9, 16), new Point(10, 16)));
        Sample s7 = Sample.createGridToGridSample(2, Arrays.asList(new Point(15, 16), new Point(16, 16)),
                Arrays.asList(new Point(13, 1), new Point(13, 2)));

        List<Sample> samples = Arrays.asList(s1, s2, s3, s4, s5, s6, s7);
        Grid grid = new Grid(16, obstacles, inputs, exits, samples);

        return new TaskContainer(grid, samples);
    }

    public static TaskContainer getBio4() {

        List<Point> obstacles = Grid.getPointsInRectangle(new Point(3, 8), new Point(5, 10));
        obstacles.addAll(Grid.getPointsInRectangle(new Point(2, 12), new Point(7, 15)));
        obstacles.addAll(Grid.getPointsInRectangle(new Point(12, 2), new Point(15, 6)));
        obstacles.addAll(Grid.getPointsInRectangle(new Point(10, 12), new Point(15, 15)));

        List<Point> inputs = Arrays.asList(new Point(1, 0), new Point(16, 17));
        List<Point> exits = Arrays.asList(new Point(16, 0), new Point(1, 17));

        Sample s1 = Sample.createGridToGridSample(2, Arrays.asList(new Point(11, 4), new Point(11, 3)),
                Arrays.asList(new Point(11, 9), new Point(11, 8)));

        List<Sample> samples = Arrays.asList(s1);
        Grid grid = new Grid(16, obstacles, inputs, exits, samples);

        return new TaskContainer(grid, samples);
    }

    public static TaskContainer getBio5() {

        List<Point> obstacles = Grid.getPointsInRectangle(new Point(1, 13), new Point(2, 15));
        obstacles.addAll(Grid.getPointsInRectangle(new Point(3, 8), new Point(5, 10)));
        obstacles.addAll(Grid.getPointsInRectangle(new Point(12, 8), new Point(14, 10)));
        obstacles.addAll(Grid.getPointsInRectangle(new Point(15, 13), new Point(16, 15)));

        List<Point> inputs = Arrays.asList(new Point(1, 0), new Point(16, 17));
        List<Point> exits = Arrays.asList(new Point(16, 0), new Point(1, 17));

        Sample s1 = Sample.createGridToGridSample(2, Arrays.asList(new Point(12, 7), new Point(13, 7)),
                Arrays.asList(new Point(8, 1), new Point(8, 2)));
        Sample s2 = Sample.createGridToGridSample(2, Arrays.asList(new Point(16, 5), new Point(16, 6)),
                Arrays.asList(new Point(9, 1), new Point(9, 2)));
        Sample s3 = Sample.createGridToGridSample(2, Arrays.asList(new Point(15, 7), new Point(16, 7)),
                Arrays.asList(new Point(8, 10), new Point(9, 10)));
        Sample s4 = Sample.createGridToGridSample(2, Arrays.asList(new Point(16, 12), new Point(16, 11)),
                Arrays.asList(new Point(8, 11), new Point(9, 11)));

        List<Sample> samples = Arrays.asList(s1, s2, s3, s4);
        Grid grid = new Grid(16, obstacles, inputs, exits, samples);

        return new TaskContainer(grid, samples);
    }

    public static TaskContainer getBio6_a() {

        List<Point> obstacles = Grid.getPointsInRectangle(new Point(7, 2), new Point(10, 6));
        obstacles.addAll(Grid.getPointsInRectangle(new Point(7, 11), new Point(10, 15)));
        obstacles.addAll(Grid.getPointsInRectangle(new Point(6, 9), new Point(7, 9)));

        List<Point> inputs = Arrays.asList(new Point(1, 0), new Point(16, 17));
        List<Point> exits = Arrays.asList(new Point(16, 0), new Point(1, 17));

        Sample s1 = Sample.createGridToGridSample(2, Arrays.asList(new Point(5, 14), new Point(5, 15)),
                Arrays.asList(new Point(2, 9), new Point(2, 10)));
        Sample s2 = Sample.createGridToGridSample(2, Arrays.asList(new Point(11, 9), new Point(12, 9)),
                Arrays.asList(new Point(8, 9), new Point(9, 9)));
        Sample s3 = Sample.createGridToGridSample(2, Arrays.asList(new Point(12, 14), new Point(12, 15)),
                Arrays.asList(new Point(15, 9), new Point(15, 10)));

        List<Sample> samples = Arrays.asList(s1, s2, s3);
        Grid grid = new Grid(16, obstacles, inputs, exits, samples);

        return new TaskContainer(grid, samples);
    }

    public static TaskContainer getBio6_b() {

        List<Point> obstacles = Grid.getPointsInRectangle(new Point(7, 2), new Point(10, 6));
        obstacles.addAll(Grid.getPointsInRectangle(new Point(7, 11), new Point(10, 15)));
        obstacles.addAll(Grid.getPointsInRectangle(new Point(2, 9), new Point(2, 10)));
        obstacles.addAll(Grid.getPointsInRectangle(new Point(15, 9), new Point(15, 10)));

        List<Point> inputs = Arrays.asList(new Point(1, 0), new Point(16, 17));
        List<Point> exits = Arrays.asList(new Point(16, 0), new Point(1, 17));

        Sample s1 = Sample.createGridToGridSample(4, Arrays.asList(new Point(6, 9), new Point(7, 9), new Point(8, 9), new Point(9, 9)),
                Arrays.asList(new Point(10, 16), new Point(9, 16), new Point(8, 16), new Point(7, 16)));

        List<Sample> samples = Arrays.asList(s1);
        Grid grid = new Grid(16, obstacles, inputs, exits, samples);

        return new TaskContainer(grid, samples);
    }

    public static TaskContainer getBio7() {

        List<Point> obstacles = Grid.getPointsInRectangle(new Point(3, 8), new Point(5, 10));
        obstacles.addAll(Grid.getPointsInRectangle(new Point(12, 8), new Point(14, 10)));

        List<Point> inputs = Arrays.asList(new Point(1, 0), new Point(16, 17));
        List<Point> exits = Arrays.asList(new Point(16, 0), new Point(1, 17));

        Sample s1 = Sample.createGridToGridSample(2, Arrays.asList(new Point(8, 7), new Point(8, 6)),
                Arrays.asList(new Point(5, 14), new Point(5, 13)));
        Sample s2 = Sample.createGridToGridSample(2, Arrays.asList(new Point(8, 10), new Point(8, 9)),
                Arrays.asList(new Point(12, 14), new Point(12, 13)));

        List<Sample> samples = Arrays.asList(s1, s2);
        Grid grid = new Grid(16, obstacles, inputs, exits, samples);

        return new TaskContainer(grid, samples);
    }

    public static TaskContainer getBio8() {

        List<Point> obstacles = Grid.getPointsInRectangle(new Point(3, 8), new Point(5, 10));
        obstacles.addAll(Grid.getPointsInRectangle(new Point(2, 13), new Point(4, 15)));

        List<Point> inputs = Arrays.asList(new Point(1, 0), new Point(16, 17));
        List<Point> exits = Arrays.asList(new Point(16, 0), new Point(1, 17));

        Sample s1 = Sample.createGridToGridSample(2, Arrays.asList(new Point(11, 9), new Point(11, 8)),
                Arrays.asList(new Point(9, 16), new Point(9, 15)));
        Sample s2 = Sample.createGridToGridSample(2, Arrays.asList(new Point(16, 14), new Point(16, 15)),
                Arrays.asList(new Point(15, 8), new Point(15, 9)));

        List<Sample> samples = Arrays.asList(s1, s2);
        Grid grid = new Grid(16, obstacles, inputs, exits, samples);

        return new TaskContainer(grid, samples);
    }

    public static TaskContainer getBio9() {

        List<Point> obstacles = Grid.getPointsInRectangle(new Point(12, 8), new Point(14, 10));

        List<Point> inputs = Arrays.asList(new Point(1, 0), new Point(16, 17));
        List<Point> exits = Arrays.asList(new Point(16, 0), new Point(1, 17));

        Sample s1 = Sample.createGridToGridSample(2, Arrays.asList(new Point(1, 14), new Point(1, 15)),
                Arrays.asList(new Point(2, 9), new Point(2, 10)));
        Sample s2 = Sample.createGridToGridSample(2, Arrays.asList(new Point(6, 9), new Point(6, 8)),
                Arrays.asList(new Point(9, 16), new Point(9, 15)));

        List<Sample> samples = Arrays.asList(s1, s2);
        Grid grid = new Grid(16, obstacles, inputs, exits, samples);

        return new TaskContainer(grid, samples);
    }

    public static TaskContainer getBio10() {

        List<Point> obstacles = Grid.getPointsInRectangle(new Point(3, 8), new Point(5, 10));

        List<Point> inputs = Arrays.asList(new Point(1, 0), new Point(16, 17));
        List<Point> exits = Arrays.asList(new Point(16, 0), new Point(1, 17));

        Sample s1 = Sample.createGridToGridSample(2, Arrays.asList(new Point(11, 9), new Point(11, 8)),
                Arrays.asList(new Point(9, 16), new Point(9, 15)));

        List<Sample> samples = Arrays.asList(s1);
        Grid grid = new Grid(16, obstacles, inputs, exits, samples);

        return new TaskContainer(grid, samples);
    }

    public static TaskContainer getBio11() {

        List<Point> obstacles = Collections.emptyList();

        List<Point> inputs = Arrays.asList(new Point(1, 0), new Point(16, 17));
        List<Point> exits = Arrays.asList(new Point(16, 0), new Point(1, 17));

        Sample s1 = Sample.createGridToGridSample(2, Arrays.asList(new Point(6, 9), new Point(6, 8)),
                Arrays.asList(new Point(9, 16), new Point(9, 15)));

        List<Sample> samples = Arrays.asList(s1);
        Grid grid = new Grid(16, obstacles, inputs, exits, samples);

        return new TaskContainer(grid, samples);
    }

    private InVitro1() {

    }
}
