/*
 * Copyright (c) 2018 Institute for Integrated Circuits, Johannes Kepler University.

 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:

 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.

 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
*/
package at.jku.iic.pmd.mazerouter;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.stream.Collectors;

import at.jku.iic.pmd.Grid;
import at.jku.iic.pmd.RoutingDescription;
import at.jku.iic.pmd.Sample;

/**
 * Executes the routing of the samples by using the external tool &quot;Multi-sink lee-moore shortest path
 * maze router&quot;.
 */
public class MazeRouterCaller {

    private static final String MAZE_ROUTER_EXECUTABLE_PATH = "../maze_router/maze";

    /**
     * For a detailed description of this algorithm we refer to our paper &quot;Sound Valve-Control for
     * Programmable Microfluidic Devices&quot;.
     * 
     */
    public boolean executeRouting(Grid grid, List<Sample> samples) throws IOException {

        Set<Sample> routed = new HashSet<>();

        int time = 0;
        while (samples.size() != routed.size() && time != -1) {

            for (Sample s : samples) {
                if (!routed.contains(s)) {

                    if (callMazeRouting(grid, s, time)) {
                        routed.add(s);
                    }
                }
            }

            time = grid.getNextTimeOfGridChange(time);
        }

        return samples.size() == routed.size();
    }

    private boolean callMazeRouting(Grid grid, Sample s, int time) throws IOException {

        final String FILE_NAME = "../tmp.infile";
        for (RoutingDescription possibleRoutingDescription : s.getRoutingDescriptions(grid, time)) {

            PrintWriter out = new PrintWriter(FILE_NAME);
            out.write(grid.getStringRep(s, time, possibleRoutingDescription.usedInput, possibleRoutingDescription.usedExit));
            out.write("\n");
            out.write(possibleRoutingDescription.stringRepresentation);
            out.close();

            Process process = new ProcessBuilder(MAZE_ROUTER_EXECUTABLE_PATH, FILE_NAME).start();
            InputStream inputStream = process.getInputStream();
            BufferedReader br = new BufferedReader(new InputStreamReader(inputStream));
            String result = br.lines().collect(Collectors.joining("\n"));
            br.close();

            if (Integer.parseInt(result.substring(0, result.indexOf('\n'))) == s.routingSegments) {
                // Routing success
                grid.updateGrid(s, time, result.substring(result.indexOf('\n') + 1));
                return true;
            }
        }
        return false;
    }
}
