/*
 * Copyright (c) 2018 Institute for Integrated Circuits, Johannes Kepler University.

 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:

 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.

 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
*/
package at.jku.iic.pmd.test;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;

import java.awt.Point;
import java.io.IOException;
import java.nio.file.Paths;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;

import org.junit.Test;

import at.jku.iic.pmd.TaskContainer;
import at.jku.iic.pmd.Grid;
import at.jku.iic.pmd.Sample;
import at.jku.iic.pmd.benchmark.InVitro1;
import at.jku.iic.pmd.benchmark.InVitro2;
import at.jku.iic.pmd.mazerouter.MazeRouterCaller;
import at.jku.iic.pmd.parser.RoutabilityDrivenBenchmarkParser;

/**
 * Integration tests.
 */
public class RoutingTests {

    @Test
    public void gridToGridRoutingTest() throws IOException {

        List<Point> obstacles = Arrays.asList(new Point(4, 3), new Point(4, 4));
        List<Point> inputs = Collections.singletonList(new Point(1, 0));
        List<Point> exits = Collections.singletonList(new Point(5, 0));

        Sample s1 = Sample.createGridToGridSample(2, Arrays.asList(new Point(3, 2), new Point(2, 2)),
                Arrays.asList(new Point(5, 3), new Point(5, 4)));

        List<Sample> samples = Arrays.asList(s1);
        Grid grid = new Grid(5, obstacles, inputs, exits, samples);

        MazeRouterCaller router = new MazeRouterCaller();
        assertTrue(router.executeRouting(grid, samples));

        assertEquals(
                "-1  7 -1 -1 -1  7 -1 \n" + "-1  7  7  0  0  7 -1 \n" + "-1  0  7  7  0  7 -1 \n" + "-1  0  0  7 -1 -1 -1 \n"
                        + "-1  0  0  7 -1 -1 -1 \n" + "-1  0  0  7  7  7 -1 \n" + "-1 -1 -1 -1 -1 -1 -1 \n",
                grid.toString(grid.getLatestTimeOfGridChange() - 1));
    }

    @Test
    public void inputToExitRoutingTest() throws IOException {

        List<Point> obstacles = Arrays.asList(new Point(4, 3), new Point(4, 4));
        List<Point> inputs = Collections.singletonList(new Point(1, 0));
        List<Point> exits = Collections.singletonList(new Point(5, 0));

        Sample s1 = Sample.createInputToExitSample(2, inputs.get(0), exits.get(0));

        List<Sample> samples = Arrays.asList(s1);
        Grid grid = new Grid(5, obstacles, inputs, exits, samples);

        MazeRouterCaller router = new MazeRouterCaller();
        assertTrue(router.executeRouting(grid, samples));

        assertEquals(
                "-1  7 -1 -1 -1  7 -1 \n" + "-1  7  7  7  7  7 -1 \n" + "-1  0  0  0  0  0 -1 \n" + "-1  0  0  0 -1  0 -1 \n"
                        + "-1  0  0  0 -1  0 -1 \n" + "-1  0  0  0  0  0 -1 \n" + "-1 -1 -1 -1 -1 -1 -1 \n",
                grid.toString(grid.getLatestTimeOfGridChange() - 1));
    }

    @Test
    public void gridToExitRoutingTest() throws IOException {

        List<Point> obstacles = Arrays.asList(new Point(4, 3), new Point(4, 4));
        List<Point> inputs = Collections.singletonList(new Point(1, 0));
        List<Point> exits = Collections.singletonList(new Point(5, 0));

        Sample s1 = Sample.createGridToExitSample(2, exits.get(0), Arrays.asList(new Point(5, 3), new Point(5, 4)));

        List<Sample> samples = Arrays.asList(s1);
        Grid grid = new Grid(5, obstacles, inputs, exits, samples);
        MazeRouterCaller router = new MazeRouterCaller();
        assertTrue(router.executeRouting(grid, samples));

        assertEquals(
                "-1  4 -1 -1 -1  4 -1 \n" + "-1  4  0  0  0  4 -1 \n" + "-1  4  0  0  0  4 -1 \n" + "-1  4  0  0 -1  4 -1 \n"
                        + "-1  4  0  0 -1  4 -1 \n" + "-1  4  4  4  4  4 -1 \n" + "-1 -1 -1 -1 -1 -1 -1 \n",
                grid.toString(grid.getLatestTimeOfGridChange() - 1));
    }

    @Test
    public void intputToGridRoutingTest() throws IOException {

        List<Point> obstacles = Arrays.asList(new Point(4, 3), new Point(4, 4));
        List<Point> inputs = Collections.singletonList(new Point(1, 0));
        List<Point> exits = Collections.singletonList(new Point(5, 0));

        Sample s1 = Sample.createInputToGridSample(2, inputs.get(0), Arrays.asList(new Point(5, 3), new Point(5, 4)));

        List<Sample> samples = Arrays.asList(s1);
        Grid grid = new Grid(5, obstacles, inputs, exits, samples);
        MazeRouterCaller router = new MazeRouterCaller();
        assertTrue(router.executeRouting(grid, samples));

        assertEquals(
                "-1 11 -1 -1 -1 11 -1 \n" + "-1 11  0  0  0 11 -1 \n" + "-1 11  0  0  0 11 -1 \n" + "-1 11  0  0 -1 -1 -1 \n"
                        + "-1 11  0  0 -1 -1 -1 \n" + "-1 11 11 11 11 11 -1 \n" + "-1 -1 -1 -1 -1 -1 -1 \n",
                grid.toString(grid.getLatestTimeOfGridChange() - 1));
    }

    @Test
    public void ripUpTest() throws IOException {

        List<Point> obstacles = Arrays.asList(new Point(8, 5), new Point(8, 6), new Point(8, 7), new Point(8, 8), new Point(8, 9));
        List<Point> inputs = Collections.singletonList(new Point(1, 0));
        List<Point> exits = Collections.singletonList(new Point(10, 0));

        Sample s1 = Sample.createGridToExitSample(2, exits.get(0), Arrays.asList(new Point(9, 5), new Point(9, 6)));

        List<Sample> samples = Arrays.asList(s1);
        Grid grid = new Grid(10, obstacles, inputs, exits, samples);
        MazeRouterCaller router = new MazeRouterCaller();
        assertTrue(router.executeRouting(grid, samples));

        assertEquals(
                "-1  7 -1 -1 -1 -1 -1 -1 -1 -1  7 -1 \n" + "-1  7  0  0  0  0  0  0  0  0  7 -1 \n"
                        + "-1  7  0  0  0  0  0  0  0  0  7 -1 \n" + "-1  7  0  0  0  0  0  0  0  0  7 -1 \n"
                        + "-1  7  0  0  0  0  0  0  0  0  7 -1 \n" + "-1  7  0  0  0  0  0  0 -1  7  7 -1 \n"
                        + "-1  7  0  0  0  0  0  0 -1  7  0 -1 \n" + "-1  7  0  0  0  0  0  0 -1  7  0 -1 \n"
                        + "-1  7  0  0  0  0  0  0 -1  7  0 -1 \n" + "-1  7  0  0  0  0  0  0 -1  7  0 -1 \n"
                        + "-1  7  7  7  7  7  7  7  7  7  0 -1 \n" + "-1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 \n",
                grid.toString(grid.getLatestTimeOfGridChange() - 1));
    }

    @Test
    public void inVitro1Test() throws IOException {

        int totalTime = 0;
        MazeRouterCaller router = new MazeRouterCaller();
        for (TaskContainer subPart : InVitro1.getBenchmark()) {

            assertTrue(router.executeRouting(subPart.getGrid(), subPart.getSamples()));
            totalTime += subPart.getGrid().getLatestTimeOfGridChange();
        }
        assertEquals(297, totalTime);
    }

    @Test
    public void inVitro2Test() throws IOException {

        int totalTime = 0;
        MazeRouterCaller router = new MazeRouterCaller();
        for (TaskContainer subPart : InVitro2.getBenchmark()) {

            assertTrue(router.executeRouting(subPart.getGrid(), subPart.getSamples()));
            totalTime += subPart.getGrid().getLatestTimeOfGridChange();
        }
        assertEquals(290, totalTime);
    }

    @Test
    public void testRoutabilityDrivenBenchmark() throws IOException {

        RoutabilityDrivenBenchmarkParser parser = new RoutabilityDrivenBenchmarkParser();
        TaskContainer benchmark = parser.readBenchmark(Paths.get("../benchmarks/case10x10_flownum5/random0"));

        MazeRouterCaller router = new MazeRouterCaller();
        assertTrue(router.executeRouting(benchmark.getGrid(), benchmark.getSamples()));
        assertEquals(58, benchmark.getGrid().getLatestTimeOfGridChange());
    }
}
